'''
Generate a simple Hydra io network for specified io channels and io groups.
In this case, a 'simple' network applies the following constraints on the network:

     - All configuration packets originate at one point (head) and all data packets
     are returned via the same point
     - Hydra nodes are arranged in a regular grid, i.e. left 1, down 1 == down 1, left 1
     - Each uart is configured s.t. mosi[i] == miso_ds[i] or miso_us[i]
     - Neighboring nodes are configured s.t. miso_us[i] == mosi[j] with j being the connected
     port to i (e.g. 0<->2 and 1<->3)

'''

import numpy as np
import networkx as nx
from bidict import bidict
import random
from copy import deepcopy

default_miso_uart_map = [3,0,1,2]
default_mosi_uart_map = [0,1,2,3]
default_usds_link_map = [2,3,0,1]
dir_table = bidict({
    (0,1): 0,
    (1,0): 3,
    (-1,0): 1,
    (0,-1): 2
    })

def edge_dir(edge):
    dx = edge[1][0] - edge[0][0]
    dy = edge[1][1] - edge[0][1]
    return (dx, dy)

def chip_id_simple(g, position):
    '''
    Generates an easy-to-read chip id based on the position (excluding 0, 1, and 255)
    Chip ids increase by increments of 1 at each node position in y and 10 at each
    node position in x (works for max of 8x8 grid)

    '''
    return (position[0]*10 + position[1]) % 253 + 2

def chip_id_position(g, position):
    '''
    Generates a chip id based on the position excluding ids 0, 1, and 255
    Chip id increase by increments of 1 at each node position in y dimension
    Chip ids are only guaranteed to be unique if less than 253 positions in
    rectangle bounding all nodes

    '''
    max_pos = max(g.nodes())
    min_pos = min(g.nodes())
    id_factor = max_pos[1] - min_pos[1] + 1
    return (position[0]*id_factor + position[1]) % 253 + 2

def generate_chip_id(g, method='chip_id_simple'):
    '''
    Generates a ``chip_id`` field on each node based a dedicated scheme

    '''
    for node in g.nodes():
        if 'chip_id' in g.nodes[node]:
            continue
        g.nodes[node]['chip_id'] = globals()[method](g, node)

def generate_digraphs_from_upstream(g, root, ext):
    '''
    Generate g_us, g_ds, g_mosi assuming g is a directed graph of the upstream
    network with root node ``root`` and an external node ``ext``

    :returns: tuple of g_us, g_ds, g_mosi

    '''
    # generate upstream miso
    g_us = nx.DiGraph()
    g_us.add_nodes_from(g.nodes())
    g_us.add_edges_from([edge for edge in g.edges()])
    g_us.nodes[root]['root'] = True
    g_us.nodes[ext]['chip_id'] = 'ext'

    # generate downstream miso
    g_ds = nx.DiGraph()
    g_ds.add_nodes_from(g_us.nodes())
    g_ds.add_edges_from([(edge[::-1]) for edge in g_us.edges()])
    g_ds.nodes[root]['root'] = True
    g_ds.nodes[ext]['chip_id'] = 'ext'

    # generate mosi
    g_mosi = nx.DiGraph()
    g_mosi.add_nodes_from(g_us.nodes())
    g_mosi.add_edges_from(g_us.edges())
    g_mosi.add_edges_from(g_ds.edges())
    g_mosi.nodes[root]['root'] = True
    g_mosi.nodes[ext]['chip_id'] = 'ext'

    return g_us, g_ds, g_mosi

def generate_2d_grid_digraph(x_slots, y_slots, *args):
    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph)
    for node in args:
        g.add_node(node)
    for x in range(-1,2):
        for y in range(-1,2):
            if abs(x) + abs(y) == 1:
                for node in args:
                    if (node[0]+x,node[1]+y) in g.nodes():
                        g.add_edge(node,(node[0]+x,node[1]+y))
                        g.add_edge((node[0]+x,node[1]+y),node)
    return g

def generate_hydra_simple_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345):
    '''
    Generates a simple hydra io networks using the simple algorithm. Roughly,
    this algorithm generates a simple hydra io network based on an upstream
    network. The upstream network is found by this algorithm:

        1. create network with all connections to neighbors

        2. starting at root node, prune all incoming connections that also have an outgoing connection

        3. iteratively prune connection with the largest n_children[head] + n_children[tail]

    There is not a unique solution and so a random seed is used to generate the network

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)

    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))

    # work forward from root node pruning out links that contribute to nodes
    # with largest fifo load
    nodes = [node for node in g.nodes() if g.in_degree(node) > 1]
    while nodes:
        edges = list(g.in_edges(nodes))
        random.shuffle(edges)

        metric = [len(nx.descendants(g, edge[1])) + len(nx.descendants(g, edge[0])) for edge in edges]
        g.remove_edge(*edges[np.argmax(metric)])
        nodes = [node for node in g.nodes() if g.in_degree(node) > 1]

    return generate_digraphs_from_upstream(g, root, ext)

def generate_hydra_simple2_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345):
    '''
    Generates a simple hydra io networks using the simple2 algorithm. Roughly,
    this algorithm generates a simple hydra io network based on an upstream
    network. The upstream network is found in the same way as the simple algorithm
    except the score metric is the connection with the fewest children.

    This generally performs the best of all the algorithms developed.

    There is not a unique solution and so a random seed is used to generate the network

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)

    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))

    # get all redundant connections
    nodes = [node for node in g.nodes() if g.in_degree(node) > 1]
    while nodes:
        edges = list(g.in_edges(nodes))
        random.shuffle(edges)

        # get edge with least number of children
        edge_metric = sorted([(len(nx.descendants(g, edge[1])), edge) for edge in edges])
        g.remove_edge(*edge_metric[0][1])
        nodes = [node for node in g.nodes() if g.in_degree(node) > 1]

    return generate_digraphs_from_upstream(g, root, ext)

def L(i, d_in, n_us):
    '''
    Helper function for calculating fifo load, representing the increase in the
    fifo value during readout period ``i``. Maximum FIFO value can be found
    by summing from ``i=-1`` to ``i=d_in-1``, inclusive.

    :param i: iterator index

    :param d_in: in degree of node in downstream network

    :param n_us: sorted list of number of downstream parents of node (0 == least upstream children) of each connection in downstream network

    :returns: ``int``

    '''
    if i == -1:
        return 1
    if i == 0:
        return d_in - 1
    if i == 1:
        return (d_in - i) * (n_us[0] - 1)
    return (d_in - i) * (n_us[i-1] - n_us[i-2])

def fifo_load(g, node):
    '''
    Calculates the "FIFO load" for a given node assuming g represents the
    upstream network, and the downstream network is equivalent to the inverse
    of the upstream network.

    The fifo load is a metric representing the relative maximum fifo value reached
    by a single chip assuming all chips trigger at the same time.

    :param g: directed graph representing upstream network

    :param node: node within ``g`` to calculate fifo load

    :returns: ``int`` of fifo load

    '''
    in_links = g.out_edges(node)
    n_inputs = len(in_links)
    n_outputs = 1
    n_upstream_nodes = sorted([len(nx.descendants(g, link[1]))+1 for link in in_links])
    return sum([L(i, n_inputs, n_upstream_nodes) for i in range(-1,n_inputs)])

def network_fifo_load_score(g):
    '''
    Creates the network fifo load score as min(fifo load) + mean(fifo load).
    Modifies the ``'load'`` attribute of each node in order to save on
    computation time. When modifying the network connection (thus the fifo load),
    the ``'load'`` attributes should be cleared (see methods ``XX_edge_and_delete_fifo_load``).

    :param g: directed graph representing upstream network

    :returns: ``float``

    '''
    fifo_loads = []
    n = 0
    sum = 0
    max = 0
    for node in g.nodes():
        if not 'load' in g.nodes[node]:
            g.nodes[node]['load'] = fifo_load(g, node)
        if g.nodes[node]['load'] > max:
            max = g.nodes[node]['load']
        n += 1
        sum += g.nodes[node]['load']
    return max + sum / n

def add_edge_and_delete_fifo_load(g, edge):
    '''
    Creates edge in network and deletes the ``'load'`` attr from all nodes upstream
    of connection.

    :param g: directed graph of upstream network

    :param edge: edge to add to network

    '''
    g.add_edge(*edge)
    nodes = [edge[0]]
    while nodes:
        for node in nodes:
            if 'load' in g.nodes[node]:
                del g.nodes[node]['load']
        nodes = list(set(edge[0] for edge in g.in_edges(nodes)))

def remove_edge_and_delete_fifo_load(g, edge):
    '''
    Removes edge from network and deletes the ``'load'`` attr from all nodes
    upstream of connection.

    :param g: directed graph of upstream network

    :param edge: edge to remove from network

    '''
    g.remove_edge(*edge)
    nodes = [edge[0]]
    while nodes:
        for node in nodes:
            if 'load' in g.nodes[node]:
                del g.nodes[node]['load']
        nodes = list(set(edge[0] for edge in g.in_edges(nodes)))

def random_prune(g):
    '''
    Simple pruning algorithm:

        1. Find all edges going into nodes with in degree > 1

        2. Randomly remove (as long as all nodes maintain in degree > 1)

    :param g: directed graph representing upstream network

    :returns: ``g``

    '''
    edges_to_prune = [edge for edge in g.in_edges() if g.in_degree(edge[1]) > 1]
    random.shuffle(edges_to_prune)
    for edge in edges_to_prune:
        if g.in_degree(edge[1]) > 1:
            g.remove_edge(*edge)
    return g

def check_peturbations(g, edges, size, n, peturbation_delta=1e-6, pruning_algorithm=random_prune, scoring_algorithm=network_fifo_load_score):
    '''
    Check to see if a perturbations of the graph linkage improves (minimizes) performance

    :param g: directed graph of upstream network

    :param edges: edges to potentially add back into network (edges that already exist in network are ignored)

    :param size: number of edges to add back into network

    :param n: number of random perturbations to apply

    :param peturbation_delta: fractional change required to continue iteration

    :param pruning_algorithm: dedicated pruning algorithm to use (should take a single upstream directed graph as an argument and return a upstream single directed graph)

    :param scoring_algorithm: algorithm used to determine performance, this is the function we are minimizing (should take a single upstream directed graph as an argument and return a number)

    :returns: best performing graph found during iterations

    '''
    return_g = g
    while True:
        print('trying {} peturbations of scale {}...'.format(n, size))
        base_performance = network_fifo_load_score(return_g) # performance of iteration graph
        current_performance = base_performance # best performance of peturbation graphs
        new_edges = [edge for edge in edges if edge not in return_g.edges()]

        for i in range(n):
            random.shuffle(new_edges)
            test_g = deepcopy(return_g)
            for edge in new_edges[:min(size,len(new_edges))]:
                add_edge_and_delete_fifo_load(test_g, edge)
            test_g = pruning_algorithm(test_g)
            test_performance = network_fifo_load_score(test_g)
            if test_performance < current_performance:
                print('improved performance to {}'.format(test_performance))
                return_g = test_g
                current_performance = test_performance
        performance_delta = abs(current_performance - base_performance)/base_performance
        if performance_delta < peturbation_delta:
            break
        base_performance = current_performance
    return return_g

def generate_hydra_min_fifo_load_beam_search_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345, peturbation_delta=1e-6, beam_width=3):
    '''
    Generates a simple hydra io networks through the following algorithm:

        1. create network with all connections between neighbors

        2. starting from root node, remove all incoming connections that have an outgoing connection

        3. iteratively prune remaing connections following:

            1. calculate a weight for each redundant connection based on the fifo load score of the network with and without the connection

            2. generate a list of potential graphs (with N-1 edges) that minimize the fifo load score

            3. keep ``beam_width`` of the best performing potential graphs for next iteration

            4. iterate until all graphs have no more redundant connections (choosing the most performant)

        4. perturb the solution N/4 times by

            1. randomly add 25% of the connections pruned in step (3.)

            2. reprune

            3. keep most performant

        5. continue to perturb until the fractional improvement drops below ``peturbation_delta``

    There is not a unique solution and so a random seed is used to generate the
    network.

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :param peturbation_delta: continue to try peturbations if network score continues to change by this much (fractional, if None, don't check peturbations)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)

    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))
    shortest_paths = deepcopy(list(g.edges()))

    def prune_networks(stack, beam_width=beam_width):
        '''
        Prunes redundant incoming network connections of an upstream directed
        graph using a beam search-like algorithm.

        :param stack: a directed graph or a list of directed graphs representing the upstream network

        :param beam_width: length of stack used during search

        :returns: directed graph with redundant incoming network connections pruned

        '''
        if not isinstance(stack, list):
            stack = [stack]
        stack.sort(key=lambda x: network_fifo_load_score(x))
        while len(stack) > beam_width:
            del stack[-1]
        completed_flag = True
        next_stack = []
        for g in stack:
            edges = [edge for edge in g.in_edges() if g.in_degree(edge[1]) > 1]
            if not edges:
                next_stack += [g]
            else:
                for idx,edge in enumerate(edges):
                    remove_edge_and_delete_fifo_load(g, edge)
                    score = network_fifo_load_score(g)
                    add_edge_and_delete_fifo_load(g, edge)
                    g.edges[edge]['weight'] = score
                min_weight = min([g.edges[edge]['weight'] for edge in edges])
                edges_to_prune = [edge for edge in edges if g.edges[edge]['weight'] == min_weight]
                if not edges_to_prune:
                    next_stack += [g]
                elif len(edges_to_prune) == 1:
                    remove_edge_and_delete_fifo_load(g, edges_to_prune[0])
                    completed_flag = False
                    next_stack += [g]
                else:
                    random.shuffle(edges_to_prune)
                    for edge in edges_to_prune:
                        next_stack += [deepcopy(g)]
                        remove_edge_and_delete_fifo_load(next_stack[-1], edge)
                        completed_flag = False
        if completed_flag:
            next_stack.sort(key=lambda x: network_fifo_load_score(x))
            return next_stack[0]
        return prune_networks(next_stack, beam_width=beam_width)

    g = prune_networks(g)

    if not peturbation_delta is None:
        size = int(max(len(shortest_paths)/4,1))
        n = int(max(len(shortest_paths)/4,1))
        g = check_peturbations(g, shortest_paths, size, n, peturbation_delta=peturbation_delta, pruning_algorithm=prune_networks, scoring_algorithm=network_fifo_load_score)

    return generate_digraphs_from_upstream(g, root, ext)

# Handle for the default configuration (use beam search and peturbations)
generate_hydra_min_fifo_load_digraphs = generate_hydra_min_fifo_load_beam_search_digraphs

def generate_hydra_min_fifo_load_base_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345, peturbation_delta=None, beam_width=1):
    '''
    Handle for the ``generate_hydra_min_fifo_load_beam_search_digraphs``
    algorithm with diffent default arguments.

    '''
    return generate_hydra_min_fifo_load_beam_search_digraphs(x_slots, y_slots, root, ext, avoid=avoid, seed=seed, peturbation_delta=peturbation_delta, beam_width=beam_width)

def greedy_tree(g, root):
    '''
    Simple tree creation network for a grid-like set of nodes:

        1. Starting a root node, make all connections possible to neighbors

        2. Iterate on children created

    Insures shortest path to each node and meets simple hydra network requirements

    :param g: directed graph to add upstream connections

    :param root: root node used to start algorithm

    :returns: ``g``

    '''
    nodes = [root]
    while nodes:
        potential_links = []
        for node in nodes:
            children = [child for child in g.nodes() \
            if ((abs(node[0]-child[0]) == 1 and abs(node[1]-child[1]) == 0) \
            or (abs(node[0]-child[0]) == 0 and abs(node[1]-child[1]) == 1)) \
            and not (child == root or child == root)]
            potential_links += [(node,child) for child in children]
        random.shuffle(potential_links)
        for link in potential_links:
            if not nx.has_path(g,root,link[1]):
                g.add_edge(*link)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))
    return g

def generate_hydra_fifo_product_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345, peturbation_delta=1e-6):
    '''
    Generates a simple hydra io networks through the following algorithm:

        1. create network with all connections between neighbors

        2. starting from root node, remove all incoming connections that have an outgoing connection

        3. create maximally spanning arborescence using fifo_load(head) * fifo_load(tail) as the weight

        4. perturb the solution N times by

            1. randomly adding 25% of the connections pruned in step (3.)

            2. reprune

            3. keep most performant

        5. continue to perturb until the fractional improvement drops below ``peturbation_delta``

    There is not a unique solution and so a random seed is used to generate the
    network.

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :param peturbation_delta: continue to try peturbations if network score continues to change by this much (fractional, if None, don't check peturbations)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''

    random.seed(seed)

    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))
    shortest_paths = deepcopy(list(g.edges()))

    # calculate fifo product
    network_fifo_load_score(g)
    for edge in g.edges():
        g.edges[edge]['weight'] = g.nodes[edge[0]]['load'] * g.nodes[edge[1]]['load']

    g = nx.maximum_spanning_arborescence(g)

    if not peturbation_delta is None:
        size = int(max(len(shortest_paths)/4,1))
        n = len(shortest_paths)
        g = check_peturbations(g, shortest_paths, size, n, peturbation_delta=peturbation_delta, pruning_algorithm=random_prune, scoring_algorithm=network_fifo_load_score)

    return generate_digraphs_from_upstream(g, root, ext)


def generate_hydra_greedy_tree_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345):
    '''
    Generates simple hydra io networks via the simple algorithm:

        1. starting a root node, make all outgoing connections to neighbors

        2. iterating on resulting children, make all outgoing connections to neighbors (as long as the in degree of the neighbor is < 1)

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)

    g = nx.DiGraph()
    for x in range(x_slots):
        for y in range(y_slots):
            g.add_node((x,y))
    g.add_node(ext)
    g.add_node(root)

    if avoid:
        g.remove_nodes_from(avoid)

    g = greedy_tree(g, root)

    return generate_digraphs_from_upstream(g, root, ext)


def generate_network_config(network_graphs, io_group, io_channel, name=None, previous=None):
    '''
    Converts a hydra network configuration graph into a json-encodable dict
    conforming to the Controller network configuration specification

    :param network_graphs: 3-tuple of networkx graph objects representing the miso_us, miso_ds, and mosi networks. Each node must have ``'chip_id'`` fields

    :param io_group: The io group for this hydra network

    :param io_channel: This io channel for this hydra network

    :param name: Name of configuration, default is ``max(nodes)``

    :param previous: A dict conforming to the Controller network configuration specification to add/overwrite

    :returns: json-ready ``dict`` conforming to Controller network configuration specification

    '''
    g = network_graphs[0] # upstream (convienence handle)
    network_config = {
        '_config_type': 'controller',
        'name': '{}'.format(max(g.nodes())),
        'asic_version': 2,
        'layout': None,
        'network': {}
        }
    if previous:
        network_config = previous
    if name:
        network_config['name'] = name
    io_group, io_channel = str(io_group), str(io_channel)
    if not io_group in network_config['network']:
        network_config['network'][io_group] = dict()
    if not io_channel in network_config['network'][io_group]:
        network_config['network'][io_group][io_channel] = dict()
    network_config['network'][io_group][io_channel]['nodes'] = []

    if not 'miso_uart_map' in network_config['network']:
        network_config['network']['miso_uart_map'] = default_miso_uart_map
    if not 'mosi_uart_map' in network_config['network']:
        network_config['network']['mosi_uart_map'] = default_mosi_uart_map
    if not 'usds_link_map' in network_config['network']:
        network_config['network']['usds_link_map'] = default_usds_link_map

    subnetwork = network_config['network'][io_group][io_channel]
    if not default_miso_uart_map == network_config['network']['miso_uart_map']:
        subnetwork['miso_uart_map'] = default_miso_uart_map
    if not default_mosi_uart_map == network_config['network']['mosi_uart_map']:
        subnetwork['mosi_uart_map'] = default_mosi_uart_map
    if not default_usds_link_map == network_config['network']['usds_link_map']:
        subnetwork['usds_link_map'] = default_usds_link_map

    next_nodes = [node for node in g.nodes() if 'root' in g.nodes[node]]
    while next_nodes:
        for node in next_nodes:
            subnetwork['nodes'] += [{
                'chip_id': g.nodes[node]['chip_id'],
                'miso_us': [None]*4
                }]
            chip_config = subnetwork['nodes'][-1]
            for child in g.successors(node):
                child_idx = dir_table[edge_dir((node,child))]
                chip_config['miso_us'][child_idx] = g.nodes[child]['chip_id']
            if not any(chip_config['miso_us']):
                del chip_config['miso_us']
            if 'root' in g.nodes[node]:
                chip_config['root'] = True
        next_nodes = [edge[1] for edge in g.out_edges(next_nodes)]
    return network_config

def plot(g, name='default', labels='chip_id', figsize=(6, 6), interactive=True):
    '''
    Draws a network configuration graph (using matplotlib)

    :param g: A networkx graph object representing the configuration, each node must be a 2-``tuple`` with a ``'chip_id'`` field

    :param name: Figure name (if `'default'` will clear plot before redrawing)

    :param labels: Attribute to label nodes by, if None, don't label nodes

    :param figsize: Size of figure to draw

    :param interactive: turn on pyplot's interactive mode

    '''
    import matplotlib.pyplot as plt
    if interactive:
        plt.ion()
    plt.figure(name, figsize=figsize)
    if name == 'figure':
        plt.cla()
    if labels:
        nx.draw_networkx(g, dict([(node,node) for node in g.nodes]), labels=dict([(node,g.nodes[node][labels]) for node in g.nodes]))
    else:
        nx.draw_networkx(g, dict([(node,node) for node in g.nodes]), labels=dict([(node,'') for node in g.nodes]))
    plt.xlabel('x index')
    plt.ylabel('y index')
    plt.tight_layout()
    if not interactive:
        plt.savefig(name+'.png')

if __name__ == '__main__':
    import argparse
    import os
    import json

    parser = argparse.ArgumentParser(usage=__doc__)
    parser.add_argument('--outfile', '-o', type=str, required=False, default=None,
        help='''
        output filename (optional, default=<name>.json)
        if file exists, io groups / io channels will be added to existing file
        ''')
    parser.add_argument('--name', '-n', type=str, required=True, help='''
        configuration name (required)
        ''')
    parser.add_argument('--seed', '-s', type=int, required=False, default=12345, help='''
        seed for random shuffles
        ''')
    parser.add_argument('--x_slots', '-x', type=int, required=True, help='''
        max number of slots for network nodes (in x dim)
        ''')
    parser.add_argument('--y_slots', '-y', type=int, required=True, help='''
        max number of slots for network nodes (in y dim)
        ''')
    parser.add_argument('--root', '-r', type=json.loads, required=False, default=None, help='''
        position for root node as list '[x,y]', (optional, default=<ext node>)
        ''')
    parser.add_argument('--ext', '-e', type=json.loads, required=True, help='''
        position for ext node as list '[x,y]', (optional, default=<ext node>)
        ''')
    parser.add_argument('--avoid', type=json.loads, required=False, default=list(), help='''
        positions of nodes to avoid, specified as list of x, y positions '[[x,y],[x,y]]'
        ''')
    parser.add_argument('--chip_id', type=str, required=False, default='chip_id_simple',
        help='''sets method for generating and assigning chip ids, options are
        'chip_id_simple' and 'chip_id_position'
        ''')
    parser.add_argument('--algorithm', type=str, required=False, default='simple2',
        help='''sets method for generating network links, options are 'min_fifo_load', 'min_fifo_load_base', 'min_fifo_load_beam_search', 'fifo_product', 'greedy_tree', 'simple', and 'simple2
        ''')
    parser.add_argument('--io_group', type=int, required=False, default=1,
        help='''sets io group
        ''')
    parser.add_argument('--io_channel', type=int, required=False, default=1,
        help='''sets io channel
        ''')
    parser.add_argument('--plot', action='store_true', help='''
        flag to plot network instead of generating file
        ''')

    args = parser.parse_args()
    outfile = args.outfile
    if not outfile:
        outfile = args.name + '.json'
    avoid = [tuple(pos) for pos in args.avoid]
    ext = tuple(args.ext)
    root = tuple(args.root) if args.root is not None else ext
    print('seed is {}'.format(args.seed))
    print('creating {}x{} network with {} starting at {}, avoiding {}...'.format(args.x_slots, args.y_slots, args.algorithm, root, avoid))
    g_us, g_ds, g_mosi = globals()['generate_hydra_' + args.algorithm + '_digraphs'](args.x_slots,
        args.y_slots, root, ext, avoid=avoid, seed=args.seed)

    print('generating chip ids...')
    generate_chip_id(g_us, method=args.chip_id) # add chip ids to each node
    generate_chip_id(g_ds, method=args.chip_id) # add chip ids to each node
    generate_chip_id(g_mosi, method=args.chip_id) # add chip ids to each node

    print('calculating metrics...')
    fifo_loads = [fifo_load(g_us, node) for node in g_us.nodes()]
    n_children = [len(nx.descendants(g_us,node)) for node in g_us.nodes()]
    print('nodes: ', len(fifo_loads))
    print('mean fifo load: ', np.mean(fifo_loads))
    print('max fifo load: ', max(fifo_loads))

    if args.plot:
        plot(g_us, name='MISO upstream')
        plot(g_ds, name='MISO downstream')
        plot(g_mosi, name='MOSI')
        print('Press enter to quit')
        input()
        exit()

    else:
        plot(g_us, name=outfile[:-5], interactive=False)

        previous_config = None
        if os.path.exists(outfile):
            print('loading existing file',outfile)
            with open(outfile,'r') as f:
                previous_config = json.load(f)
        new_config = generate_network_config((g_us, g_ds, g_mosi), args.io_group, args.io_channel, name=args.name, previous=previous_config)
        with open(outfile,'w') as f:
            json.dump(new_config, f, indent=4)



from larpix.format.pacman_msg_format import *
from larpix import Packet_v2, TimestampPacket, SyncPacket, TriggerPacket

def test_header():
    print('test_header')
    for i,msg_type in enumerate(msg_type_table):
        bs = format_header(msg_type,i)
        print(bs)
        prsd = parse_header(bs)
        assert prsd[0] == msg_type
        assert prsd[-1] == i

def test_word():
    print('test_word')
    test_data = dict(
        DATA = (1,2,b'testing!'),
        TRIG = (b'\x01',2),
        SYNC = (b'H',1,2),
        PING = (),
        WRITE = (1,2),
        READ = (1,2),
        TX = (1,b'testing!'),
        PONG = (),
        ERR = (1,b'123456789ABCDE')
        )
    for msg_type in msg_type_table:
        for word_type in word_type_table[msg_type]:
            bs = format_word(msg_type, word_type, *test_data[word_type])
            print(bs)
            prsd = parse_word(msg_type, bs)
            assert prsd[0] == word_type
            assert prsd[1:] == test_data[word_type]

def test_packets():
    packets = []
    packets.append(TimestampPacket(timestamp=1234567))
    for i in range(100):
        packets.append(Packet_v2())
        packets[-1].io_channel = i
    packets.append(SyncPacket(timestamp=123456, sync_type=b'H', clk_source=1))
    packets.append(TriggerPacket(timestamp=123456, trigger_type=b'\x01'))

    msg = format(packets, msg_type='DATA')
    print(msg)
    new_packets = parse(msg)

    assert packets[1:] == new_packets[1:]
    assert isinstance(new_packets[0], TimestampPacket)



import json
import os

def _load_and_verify(file, config_type=None):
    '''
    Loads full inheritance and checks that configuration type matches type requested

    '''
    config_data = json.load(file)
    if config_type:
        assert config_data['_config_type'] == config_type, "Invalid config type {}".format(config_data['_config_type'])
    return _load_inheritance(config_data)

def _inherit_dict(d, other, config_type=None):
    '''
    Copies other and overwrites with values from d following inheritance rules
    of the config_type

    '''
    combined_d = dict(other.items())
    for key,val in d.items():
        if isinstance(val, dict) and key in other:
            combined_d[key] = _inherit_dict(val, other[key], config_type)
        else:
            combined_d[key] = val
    return combined_d

def _load_inheritance(config):
    '''
    Loads parent config files specified in the "_include" field of a
    configuration.

    All files must be of the same config type and are loaded in the order
    specified in the list.

    All fields of dict-like objects within the included fields will be inherited
    recursively

    '''

    if not '_include' in config:
        return config

    base_config = dict()
    if not isinstance(config['_include'],list):
        raise RuntimeError('inherited files not specified as list')
    config_type = None
    if '_config_type' in config:
        config_type = config['_config_type']
    for included_file in config['_include']:
        inherited_config = load(included_file, config_type=config_type)
        for key,val in inherited_config.items():
            if isinstance(val, dict) and key in base_config:
                base_config[key] = _inherit_dict(val, base_config[key], config_type)
            else:
                base_config[key] = val
    for key,val in config.items():
        if isinstance(val, dict) and key in base_config:
            base_config[key] = _inherit_dict(val, base_config[key], config_type)
        else:
            base_config[key] = val
    return base_config

def load(filename, config_type=None):
    '''
    Load the specified configuration file.

    The path is first searched relative to the "current directory".
    If no match is found, the path is searched relative to the "config"
    package directory (aka ``__file__`` directly in code).

    '''
    if os.path.isfile(filename):
        with open(filename, 'r') as f:
            data = _load_and_verify(f, config_type)
            return data
    elif os.path.isfile(os.path.join(os.path.dirname(__file__), filename)):
        with open(os.path.join(os.path.dirname(__file__), filename), 'r') as f:
            return _load_and_verify(f, config_type)
    else:
        raise IOError('File not found: %s' % filename)



'''
The module contains all the message formats used by systems that interface
with larpix-control:

    - dataserver_message_encode: convert from packets to the data server messaging format
    - dataserver_message_decode: convert from data server messaging format to packets

'''
import warnings
import struct

from larpix.larpix import Packet, TimestampPacket, Packet_v1, Packet_v2

def dataserver_message_encode(packets, version=(1,0)):
    r'''
    Convert a list of packets to larpix dataserver messages. DAQ board messages are formatted using 8-byte words with the first word being a header word describing the interpretation of other words in the message. These messages are formatted as follows

        All messages:
         - byte[0] = major version
         - byte[1] = minor version
         - byte[2] = message type

            - ``b'D'``: LArPix data
            - ``b'T'``: Timestamp data
            - ``b'H'``: Heartbeat

        LArPix heartbeat messages:
         - byte[3] = ``b'H'``
         - byte[4] = ``b'B'``
         - byte[5:7] are unused

        LArPix data messages:
         - byte[3] = io channel
         - bytes[4:7] are unused
         - bytes[8:] = raw LArPix 7-byte UART words ending in a single null byte

        Timestamp data messages:
         - byte[3:7] are unused
         - byte[8:14] = 7-byte Unix timestamp
         - byte[15] is unused

    Example usage::

        from larpix.larpix import Packet, Key
        packet = Packet()
        packet.chip_key = Key('1-1-1')
        msgs = datserver_message_encode([packet])
        msgs[0] # b'\x01\x00D\x01\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00'
        msgs[0][:8] # header
        msgs[0][8:] # data words

    :param packets: list of ``larpix.Packet`` objects

    :param version: optional, encode a message in specified version format, ``tuple`` of major, minor numbers

    :returns: list of bytestream messages, 1 for each packet

    '''
    data_header_fmt='BBcBI'
    timestamp_header_fmt='BBcBI'
    msgs = []
    for packet in packets:
        msg = b''
        if isinstance(packet, Packet):
            header = [0]*len(data_header_fmt)
            header[0:2] = version[0:2]
            header[2] = b'D'
            if packet.io_channel:
                header[3] = packet.io_channel
            else:
                raise ValueError('all packets must have a declared io_channel')
            msg = struct.pack(data_header_fmt, *header)
            if isinstance(packet, Packet_v1):
                msg += packet.bytes() + struct.pack('B',0)
            elif isinstance(packet, Packet_v2):
                msg += packet.bytes()
        elif isinstance(packet, TimestampPacket):
            header = [0]*len(timestamp_header_fmt)
            header[0:2] = version[0:2]
            header[2] = b'T'
            msg = struct.pack(data_header_fmt, *header)
            msg += packet.bytes() + b'\x00'
        msgs += [msg]
    return msgs

def dataserver_message_decode(msgs, version=(1,0), **kwargs):
    r'''
    Convert a list of larpix data server messages into packets. Additional packet meta data can be passed along via kwargs E.g.::

        msg = b'\x01\x00D\x01\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00'
        packets = dataserver_message_decode([msg], io_group=1)
        packets[0] # Packet(b'\x04\x00\x00\x00\x00\x00\x00'), key of '1-1-1'

    :param msgs: list of bytestream messages each starting with a single 8-byte header word, followed by N 8-byte data words

    :param version: optional, message version to validate against, ``tuple`` of major, minor version numbers

    :returns: list of ``larpix.Packet`` and ``larpix.TimestampPacket`` objects

    '''
    packets = []
    for msg in msgs:
        major, minor = struct.unpack('BB',msg[:2])
        if (major, minor) != version:
            warnings.warn('Message version mismatch! Expected {}, received {}'.format(version, (major,minor)))
        msg_type = struct.unpack('c',msg[2:3])[0]
        if msg_type == b'T':
            timestamp = struct.unpack('L',msg[8:15] + b'\x00')[0] # only use 7-bytes
            packets.append(TimestampPacket(timestamp=timestamp))
            if kwargs:
                for key,value in kwargs.items():
                    setattr(packets[-1], key, value)
        elif msg_type == b'D':
            io_chain = struct.unpack('B',msg[3:4])[0]
            payload = msg[8:]
            if len(payload)%8 == 0:
                for start_index in range(0, len(payload), 8):
                    packet_bytes = payload[start_index:start_index+8]
                    if Packet == Packet_v1:
                        packets.append(Packet(packet_bytes[:-1]))
                    elif Packet == Packet_v2:
                        packets.append(Packet(packet_bytes))
                    packets[-1].io_channel = io_chain
                    if kwargs:
                        for key,value in kwargs.items():
                            setattr(packets[-1], key, value)
        elif msg_type == b'H':
            print('Heartbeat message: {}'.format(msg[3:]))
    return packets



from bitarray import bitarray
import os
import errno
import functools
from collections import OrderedDict

from .. import bitarrayhelper as bah
from .. import configs
from . import BaseConfiguration_v2, _Smart_List
from . import configuration_v2_base as v2_base


class Configuration_v2b(BaseConfiguration_v2):
    '''
    Represents the desired configuration state of a LArPix v2b chip.

    Each register name is available as its own attribute for inspecting and
    setting the value of the corresponding register.

    Certain configuration values are set channel-by-channel. These are
    represented by a list of values. For example:

        >>> conf.pixel_trim_dac[2:5]
        [16, 16, 16]
        >>> conf.channel_mask[20] = 1
        >>> conf.external_trigger_mask = [0] * 64

    Additionally, other configuration values take up more than or less
    than one complete register. These are still set by referencing the
    appropriate name. For example, ``cross_trigger_mode`` shares a
    register with a few other values, and adjusting the value of the
    ``cross_trigger_mode`` attribute will leave the other values
    unchanged.

    Each register name can cover more than one 'physical' register depending on
    the size of the data it holds. You can see which physical registers a
    given register name corresponds to by using the `register_map` attribute, e.g.::

        >>> conf.register_map['digital_threshold']  # 64 registers, 1 per channel
        range(173, 237)
        >>> conf.register_map['enable_dynamic_reset']  # Register 170
        range(170, 171)
        >>> conf.register_map['enable_min_delta_adc']  # Shares register 170
        range(170, 171)

    '''

    asic_version = '2b'
    default_configuration_file = 'chip/default_v2b.json'
    num_registers = 256
    num_bits = 2048

    # Additional class properties regarding configuration registers are set at the end of the file.

    def __init__(self):
        # Note: properties, getters and setters are constructed after this class definition at the bottom of the file.
        super(Configuration_v2b, self).__init__()
        return

## Set up property info
#
_property_configuration = OrderedDict([
        ('pixel_trim_dac',
            (v2_base._list_property, (int, 0, 31, Configuration_v2b.num_channels, 8), (0,512))),
        ('threshold_global',
            (v2_base._basic_property, (int, 0, 255), (512, 520))),
        ('csa_gain',
            (v2_base._compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (520,521))),
        ('csa_bypass_enable',
            (v2_base._compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (521,522))),
        ('bypass_caps_en',
            (v2_base._compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (522,523))),
        ('csa_enable',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2b.num_channels, 1), (528, 592))),
        ('ibias_tdac',
            (v2_base._basic_property, (int, 0, 15), (592, 596))),
        ('ibias_comp',
            (v2_base._basic_property, (int, 0, 15), (600, 604))),
        ('ibias_buffer',
            (v2_base._basic_property, (int, 0, 15), (608, 612))),
        ('ibias_csa',
            (v2_base._basic_property, (int, 0, 15), (616, 620))),
        ('ibias_vref_buffer',
            (v2_base._basic_property, (int, 0, 15), (624, 628))),
        ('ibias_vcm_buffer',
            (v2_base._basic_property, (int, 0, 15), (632, 636))),
        ('ibias_tpulse',
            (v2_base._basic_property, (int, 0, 15), (640, 644))),
        ('ref_current_trim',
            (v2_base._compound_property, (['ref_current_trim','override_ref','ref_kickstart'], int, 0, 31), (648, 653))),
        ('override_ref',
            (v2_base._compound_property, (['ref_current_trim','override_ref','ref_kickstart'], (int,bool), 0, 1), (653, 654))),
        ('ref_kickstart',
            (v2_base._compound_property, (['ref_current_trim','override_ref','ref_kickstart'], (int,bool), 0, 1), (654, 655))),
        ('vref_dac',
            (v2_base._basic_property, (int, 0, 255), (656, 664))),
        ('vcm_dac',
            (v2_base._basic_property, (int, 0, 255), (664,672))),
        ('csa_bypass_select',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2b.num_channels, 1), (672,736))),
        ('csa_monitor_select',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2b.num_channels, 1), (736,800))),
        ('csa_testpulse_enable',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2b.num_channels, 1), (800,864))),
        ('csa_testpulse_dac',
            (v2_base._basic_property, (int, 0, 255), (864,872))),
        ('current_monitor_bank0',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (872,876))),
        ('current_monitor_bank1',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (880,884))),
        ('current_monitor_bank2',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (888,892))),
        ('current_monitor_bank3',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (896,900))),
        ('voltage_monitor_bank0',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (904,907))),
        ('voltage_monitor_bank1',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (912,915))),
        ('voltage_monitor_bank2',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (920,923))),
        ('voltage_monitor_bank3',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (928,931))),
        ('voltage_monitor_refgen',
            (v2_base._list_property, ((int,bool), 0, 1, 8, 1), (936,944))),
        ('digital_monitor_enable',
            (v2_base._compound_property, (['digital_monitor_enable','digital_monitor_select'], (int,bool), 0, 1), (944,945))),
        ('digital_monitor_select',
            (v2_base._compound_property, (['digital_monitor_enable','digital_monitor_select'], (int,bool), 0, 10), (945,949))),
        ('digital_monitor_chan',
            (v2_base._basic_property, (int, 0, 63), (952,958))),
        ('adc_hold_delay',
            (v2_base._basic_property, (int, 0, 65535), (960, 976))),
        ('chip_id',
            (v2_base._basic_property, (int, 0, 255), (976, 984))),
        ('enable_tx_dynamic_powerdown',
            (v2_base._compound_property, (['enable_tx_dynamic_powerdown', 'load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl', 'tx_dynamic_powerdown_cycles'], (int, bool), 0, 1), (984, 985))),
        ('load_config_defaults',
            (v2_base._compound_property, (['enable_tx_dynamic_powerdown', 'load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl', 'tx_dynamic_powerdown_cycles'], (int, bool), 0, 1), (985, 986))),
        ('enable_fifo_diagnostics',
            (v2_base._compound_property, (['enable_tx_dynamic_powerdown', 'load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl', 'tx_dynamic_powerdown_cycles'], (int, bool), 0, 1), (986, 987))),
        ('clk_ctrl',
            (v2_base._compound_property, (['enable_tx_dynamic_powerdown', 'load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl', 'tx_dynamic_powerdown_cycles'], (int, bool), 0, 3), (987, 989))),
        ('tx_dynamic_powerdown_cycles',
            (v2_base._compound_property, (['enable_tx_dynamic_powerdown', 'load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl', 'tx_dynamic_powerdown_cycles'], (int, bool), 0, 7), (989, 992))),
        ('enable_piso_upstream',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (992, 996))),
        ('enable_piso_downstream',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (1000, 1004))),
        ('enable_posi',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (1008, 1012))),
        ('test_mode_uart0',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 3), (1016, 1018))),
        ('test_mode_uart1',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 3), (1018, 1020))),
        ('test_mode_uart2',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 3), (1020, 1022))),
        ('test_mode_uart3',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 3), (1022, 1024))),
        ('enable_cross_trigger',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1024, 1025))),
        ('enable_periodic_reset',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1025, 1026))),
        ('enable_rolling_periodic_reset',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1026, 1027))),
        ('enable_periodic_trigger',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1027, 1028))),
        ('enable_rolling_periodic_trigger',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1028, 1029))),
        ('enable_periodic_trigger_veto',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1029, 1030))),
        ('enable_hit_veto',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], int, 0, 1), (1030, 1031))),
        ('shadow_reset_length',
            (v2_base._basic_property, (int, 0, 255), (1032, 1040))),
        ('adc_burst_length',
            (v2_base._basic_property, (int, 0, 255), (1040, 1048))),
        ('channel_mask',
            (v2_base._list_property, (int, 0, 1, Configuration_v2b.num_channels, 1), (1048, 1112))),
        ('external_trigger_mask',
            (v2_base._list_property, (int, 0, 1, Configuration_v2b.num_channels, 1), (1112, 1176))),
        ('cross_trigger_mask',
            (v2_base._list_property, (int, 0, 1, Configuration_v2b.num_channels, 1), (1176, 1240))),
        ('periodic_trigger_mask',
            (v2_base._list_property, (int, 0, 1, Configuration_v2b.num_channels, 1), (1240, 1304))),
        ('periodic_reset_cycles',
            (v2_base._basic_property, (int, 0, 2**24-1), (1304, 1328))),
        ('periodic_trigger_cycles',
            (v2_base._basic_property, (int, 0, 2**32-1), (1328, 1360))),
        ('enable_dynamic_reset',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1360,1361))),
        ('enable_min_delta_adc',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1361,1362))),
        ('threshold_polarity',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1362,1363))),
        ('reset_length',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int), 0, 7), (1363,1366))),
        ('mark_first_packet',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1366,1367))),
        ('reset_threshold',
            (v2_base._basic_property, (int, 0, 255), (1368,1376))),
        ('min_delta_adc',
            (v2_base._basic_property, (int, 0, 255), (1376,1384))),
        ('digital_threshold',
            (v2_base._list_property, (int, 0, 255, Configuration_v2b.num_channels, 8), (1384,1896))),
        ('RESERVED',
            (v2_base._basic_property, (int, 0, 0), (1896, 1912))),
        ('tx_slices0',
            (v2_base._compound_property, (['tx_slices0', 'tx_slices1'], int, 0, 15), (1912, 1916))),
        ('tx_slices1',
            (v2_base._compound_property, (['tx_slices0', 'tx_slices1'], int, 0, 15), (1916, 1920))),
        ('tx_slices2',
            (v2_base._compound_property, (['tx_slices2', 'tx_slices3'], int, 0, 15), (1920, 1924))),
        ('tx_slices3',
            (v2_base._compound_property, (['tx_slices2', 'tx_slices3'], int, 0, 15), (1924, 1928))),
        ('i_tx_diff0',
            (v2_base._compound_property, (['i_tx_diff0', 'i_tx_diff1'], int, 0, 15), (1928, 1932))),
        ('i_tx_diff1',
            (v2_base._compound_property, (['i_tx_diff0', 'i_tx_diff1'], int, 0, 15), (1932, 1936))),
        ('i_tx_diff2',
            (v2_base._compound_property, (['i_tx_diff2', 'i_tx_diff3'], int, 0, 15), (1936, 1940))),
        ('i_tx_diff3',
            (v2_base._compound_property, (['i_tx_diff2', 'i_tx_diff3'], int, 0, 15), (1940, 1944))),
        ('i_rx0',
            (v2_base._compound_property, (['i_rx0', 'i_rx1'], int, 0, 15), (1944, 1948))),
        ('i_rx1',
            (v2_base._compound_property, (['i_rx0', 'i_rx1'], int, 0, 15), (1948, 1952))),
        ('i_rx2',
            (v2_base._compound_property, (['i_rx2', 'i_rx3'], int, 0, 15), (1952, 1956))),
        ('i_rx3',
            (v2_base._compound_property, (['i_rx2', 'i_rx3'], int, 0, 15), (1956, 1960))),
        ('i_rx_clk',
            (v2_base._compound_property, (['i_rx_clk', 'i_rx_rst'], int, 0, 15), (1960, 1964))),
        ('i_rx_rst',
            (v2_base._compound_property, (['i_rx_clk', 'i_rx_rst'], int, 0, 15), (1964, 1968))),
        ('i_rx_ext_trig',
            (v2_base._basic_property, (int, 0, 15), (1968, 1972))),
        ('r_term0',
            (v2_base._basic_property, (int, 0, 31), (1976, 1981))),
        ('r_term1',
            (v2_base._basic_property, (int, 0, 31), (1984, 1989))),
        ('r_term2',
            (v2_base._basic_property, (int, 0, 31), (1992, 1997))),
        ('r_term3',
            (v2_base._basic_property, (int, 0, 31), (2000, 2005))),
        ('r_term_clk',
            (v2_base._basic_property, (int, 0, 31), (2008, 2013))),
        ('r_term_reset',
            (v2_base._basic_property, (int, 0, 31), (2016, 2021))),
        ('r_term_ext_trig',
            (v2_base._basic_property, (int, 0, 31), (2024, 2029))),
        ('v_cm_lvds_tx0',
            (v2_base._compound_property, (['v_cm_lvds_tx0', 'v_cm_lvds_tx0'], int, 0, 7), (2032, 2035))),
        ('v_cm_lvds_tx1',
            (v2_base._compound_property, (['v_cm_lvds_tx0', 'v_cm_lvds_tx0'], int, 0, 7), (2036, 2039))),
         ('v_cm_lvds_tx2',
            (v2_base._compound_property, (['v_cm_lvds_tx2', 'v_cm_lvds_tx3'], int, 0, 7), (2040, 2043))),
        ('v_cm_lvds_tx3',
            (v2_base._compound_property, (['v_cm_lvds_tx2', 'v_cm_lvds_tx3'], int, 0, 7), (2044, 2047))),

    ])

# GENERATE THE PROPERTIES!
Configuration_v2b.bit_map = OrderedDict()
Configuration_v2b.register_map = OrderedDict()
Configuration_v2b.register_names = []
v2_base._generate_properties(Configuration_v2b, _property_configuration, verbose=True)



'''
Use the pytest framework to write tests for the larpix module.

'''
from __future__ import print_function
import pytest
from larpix import (Chip, Packet_v1, Packet_v2, Packet, Key, Configuration, Configuration_v1, Controller,
        PacketCollection, _Smart_List, TimestampPacket, MessagePacket)
from larpix.io import FakeIO
#from bitstring import BitArray
from bitarray import bitarray
import larpix.bitarrayhelper as bah
import json
import os

@pytest.fixture
def timestamp_packet():
    return TimestampPacket(123456789)

@pytest.fixture
def message_packet():
    return MessagePacket('test',123456789)

def list_of_packets_str(packets):
    return '\n'.join(map(str, packets)) + '\n'

def test_chip_str():
    key = '1-1-1'
    chip = Chip(key)
    result = str(chip)
    expected = 'Chip (key: 1-1-1, version: 2)'
    assert result == expected

def test_chip_get_configuration_packets(chip):
    packet_type = Packet.CONFIG_WRITE_PACKET
    packets = chip.get_configuration_write_packets()
    # test a sampling of the configuration packets
    packet = packets[5]
    assert packet.packet_type == packet_type
    assert packet.chip_id == chip.chip_id
    assert packet.register_address == 5
    assert packet.register_data == 16

    packet = packets[40]
    assert packet.packet_type == packet_type
    assert packet.chip_id == chip.chip_id
    assert packet.register_address == 40
    assert packet.register_data == 16

def test_chip_sync_configuration(chip):
    packet_type = Packet.CONFIG_READ_PACKET
    packets = chip.get_configuration_read_packets()
    chip.reads.append(PacketCollection(packets))
    chip.sync_configuration()
    result = chip.config.all_data()
    expected = [bitarray([0]*8)] * Configuration.num_registers
    assert result == expected

def test_chip_sync_configuration_slice(chip):
    packets = chip.get_configuration_read_packets()
    chip.reads.append(PacketCollection(packets[:10]))
    chip.reads.append(PacketCollection(packets[10:]))
    chip.sync_configuration(index=slice(None, None, None))
    result = chip.config.all_data()
    expected = [bitarray([0]*8)] * Configuration.num_registers
    assert result == expected

def test_chip_export_reads(chip):
    packet = Packet()
    packet.chip_key = chip.chip_key
    packet.packet_type = Packet.CONFIG_WRITE_PACKET
    packet.chip_id = chip.chip_id
    packet.register_address = 10
    packet.register_data = 20
    packet.assign_parity()
    chip.reads.append(packet)
    result = chip.export_reads()
    expected = {
            'chip_key': chip.chip_key,
            'chip_id': chip.chip_id,
            'packets': [
                packet.export()
                ]
            }
    assert result == expected
    assert chip.new_reads_index == 1

def test_chip_export_reads_no_new_reads(chip):
    result = chip.export_reads()
    expected = {'chip_id': chip.chip_id, 'chip_key': chip.chip_key, 'packets': []}
    assert result == expected
    assert chip.new_reads_index == 0
    packet = Packet()
    packet.packet_type = Packet.CONFIG_WRITE_PACKET
    chip.reads.append(packet)
    chip.export_reads()
    result = chip.export_reads()
    assert result == expected
    assert chip.new_reads_index == 1

def test_chip_export_reads_all(chip):
    packet = Packet()
    packet.chip_key = chip.chip_key
    packet.chipid = chip.chip_id
    packet.packet_type = Packet.CONFIG_WRITE_PACKET
    chip.reads.append(packet)
    chip.export_reads()
    result = chip.export_reads(only_new_reads=False)
    expected = {
            'chip_id': chip.chip_id,
            'chip_key': chip.chip_key,
            'packets': [
                packet.export()
                ]
            }
    assert result == expected
    assert chip.new_reads_index == 1

def test_controller_save_output(tmpdir, chip):
    controller = Controller()
    p = Packet()
    controller.add_chip(chip.chip_key)
    collection = PacketCollection([p], p.bytes(), 'hi', 0)
    controller.reads.append(collection)
    name = str(tmpdir.join('test.json'))
    controller.save_output(name, 'this is a test')
    with open(name) as f:
        result = json.load(f)
    expected = {
            'chips': [repr(chip)],
            'message': 'this is a test',
            'reads': [
                collection.to_dict()
                ]
            }
    assert result == expected

def test_packet_bits_bytes():
    assert Packet.num_bytes == Packet.size // 8 + min(Packet.size % 8, 1)

def test_packet_init_default():
    p = Packet()
    expected = bitarray([0] * Packet.size)
    assert p.bits == expected

def test_packet_init_bytestream():
    bytestream = b'\x3f' + b'\x00' * (Packet.num_bytes-2) + b'\xf8'
    p = Packet(bytestream)
    if isinstance(p, Packet_v1):
        expected = bitarray([0] * Packet.size)
        expected[-6:] = bitarray([1]*6)
        expected[:5] = bitarray([1]*5)
        print(expected[0],expected[-1])
        assert p.bits == expected
    else:
        expected = bitarray([0] * Packet.size)
        expected[:6] = bitarray([1]*6)
        expected[-5:] = bitarray([1]*5)
        print(expected[0],expected[-1])
        assert p.bits == expected

def test_packet_bytes_zeros():
    p = Packet()
    b = p.bytes()
    expected = b'\x00' * Packet.num_bytes
    assert b == expected

def test_packet_bytes_custom():
    p = Packet()
    p.bits[-6:] = bitarray([1]*6)
    p.bits[:7] = bitarray([0]*2+[1]*5)
    b = p.bytes()
    if isinstance(p, Packet_v1):
        expected = b'\x3f' + b'\x00' * (Packet.num_bytes-2) + b'\x3e'
        assert b == expected
    else:
        expected = b'\x7c' + b'\x00' * (Packet.num_bytes-2) + b'\xfc'
        assert b == expected

def test_packet_v1_bytes_properties():
    p = Packet_v1()
    p.packet_type = Packet_v1.DATA_PACKET
    p.chipid = 100
    expected = b'\x90\x01' + b'\x00' * (Packet_v1.num_bytes-2)
    b = p.bytes()
    assert b == expected

def test_packet_v1_export_test():
    p = Packet_v1()
    p.packet_type = Packet_v1.TEST_PACKET
    p.chipid = 5
    p.test_counter = 32838
    p.assign_parity()
    result = p.export()
    expected = {
            'asic_version': 1,
            'bits': p.bits.to01(),
            'type_str': 'test',
            'type': 1,
            'chipid': 5,
            'chip_key': None,
            'counter': 32838,
            'parity': p.parity_bit_value,
            'valid_parity': True,
            }
    assert result == expected

def test_packet_v1_export_data():
    p = Packet_v1()
    p.packet_type = Packet_v1.DATA_PACKET
    p.chipid = 2
    p.chip_key = Key('1-3-2')
    p.channel_id = 10
    p.timestamp = 123456
    p.dataword = 180
    p.fifo_half_flag = True
    p.fifo_full_flag = False
    p.assign_parity()
    result = p.export()
    expected = {
            'asic_version': 1,
            'bits': p.bits.to01(),
            'type_str': 'data',
            'type': 0,
            'chipid': 2,
            'chip_key': p.chip_key,
            'channel': 10,
            'timestamp': 123456,
            'adc_counts': 180,
            'fifo_half': True,
            'fifo_full': False,
            'parity': p.parity_bit_value,
            'valid_parity': True,
            }
    assert result == expected

def test_packet_v1_export_config_read():
    p = Packet_v1()
    p.packet_type = Packet_v1.CONFIG_READ_PACKET
    p.chipid = 10
    p.chip_key = Key('2-1-10')
    p.register_address = 51
    p.register_data = 2
    p.assign_parity()
    result = p.export()
    expected = {
            'asic_version': 1,
            'bits': p.bits.to01(),
            'type_str': 'config read',
            'type': 3,
            'chipid': 10,
            'chip_key': p.chip_key,
            'register': 51,
            'value': 2,
            'parity': p.parity_bit_value,
            'valid_parity': True,
            }
    assert result == expected

def test_packet_v1_export_config_write():
    p = Packet_v1()
    p.packet_type = Packet_v1.CONFIG_WRITE_PACKET
    p.chipid = 10
    p.register_address = 51
    p.register_data = 2
    p.assign_parity()
    result = p.export()
    expected = {
            'asic_version': 1,
            'bits': p.bits.to01(),
            'type_str': 'config write',
            'type': 2,
            'chipid': 10,
            'chip_key': p.chip_key,
            'register': 51,
            'value': 2,
            'parity': p.parity_bit_value,
            'valid_parity': True,
            }
    assert result == expected

def test_packet_v1_from_dict():
    p = Packet_v1()
    p1 = Packet_v1()
    p.packet_type = Packet_v1.CONFIG_WRITE_PACKET
    p.chipid = 10
    p.register_address = 51
    p.register_data = 2
    p.assign_parity()
    packet_dict = {
            'asic_version': 1,
            'bits': p.bits.to01(),
            'type_str': 'config write',
            'type': 2,
            'chipid': 10,
            'chip_key': None,
            'register': 51,
            'value': 2,
            'parity': p.parity_bit_value,
            'valid_parity': True,
            }
    p1.from_dict(packet_dict)
    assert p == p1

def test_packet_v1_from_dict_export_inv():
    p = Packet_v1()
    p1 = Packet_v1()
    p.packet_type = Packet_v1.CONFIG_WRITE_PACKET
    p.chipid = 10
    p.register_address = 51
    p.register_data = 2
    p.assign_parity()
    packet_dict = p.export()
    p1.from_dict(packet_dict)
    assert p == p1

def test_packet_v1_set_packet_type():
    p = Packet_v1()
    p.packet_type = Packet_v1.DATA_PACKET
    packet_bits = p.bits[Packet.packet_type_bits]
    expected = Packet_v1.DATA_PACKET
    assert packet_bits == expected

def test_packet_get_packet_type():
    p = Packet()
    p.packet_type = Packet.CONFIG_WRITE_PACKET
    packet_type = p.packet_type
    expected = Packet.CONFIG_WRITE_PACKET
    assert packet_type == expected

def test_packet_v1_set_chipid():
    p = Packet_v1()
    p.chipid = 121
    expected = bah.fromuint(121, 8)
    assert p.bits[Packet_v1.chipid_bits] == expected

def test_packet_get_chipid():
    p = Packet()
    p.chipid = 18
    expected = 18
    assert p.chipid == expected

def test_packet_v1_set_parity_bit_value():
    p = Packet_v1()
    p.parity_bit_value = 0
    assert p.bits[0] == False
    p.parity_bit_value = 1
    assert p.bits[0] == True

def test_packet_get_parity_bit_value():
    p = Packet()
    p.parity_bit_value = 0
    assert p.parity_bit_value == 0
    p.parity_bit_value = 1
    assert p.parity_bit_value == 1

def test_packet_v1_compute_parity():
    p = Packet_v1()
    p.chipid = 121
    parity = p.compute_parity()
    expected = 0
    assert parity == expected
    p.bits = bitarray([0]*54)
    parity = p.compute_parity()
    expected = 1
    assert parity == expected
    p.bits = bitarray([1]*54)
    parity = p.compute_parity()
    expected = 0
    assert parity == expected

def test_packet_v1_assign_parity():
    p = Packet_v1()
    p.chipid = 121
    p.assign_parity()
    expected = 0
    assert p.parity_bit_value == expected
    p.chipid = 0
    p.assign_parity()
    expected = 1
    assert p.parity_bit_value == expected

def test_packet_v1_has_valid_parity():
    p = Packet_v1()
    result = p.has_valid_parity()
    expected = False
    assert result == expected
    p.assign_parity()
    result = p.has_valid_parity()
    expected = True
    assert result == expected
    p.bits = bitarray([1]*54)
    result = p.has_valid_parity()
    expected = False
    assert result == expected
    p.assign_parity()
    result = p.has_valid_parity()
    expected = True
    assert result == expected

def test_packet_v1_set_channel_id():
    p = Packet_v1()
    p.channel_id = 100
    expected = bah.fromuint(100, 7)
    assert p.bits[Packet_v1.channel_id_bits] == expected

def test_packet_v1_get_channel_id():
    p = Packet_v1()
    expected = 101
    p.channel_id = expected
    assert p.channel_id == expected

def test_packet_v1_set_timestamp():
    p = Packet_v1()
    p.timestamp = 0x1327ab
    expected = bah.fromuint(int('0x1327ab', 16), 24)
    assert p.bits[Packet_v1.timestamp_bits] == expected

def test_packet_get_timestamp():
    p = Packet()
    expected = 0xa82b6e
    p.timestamp = expected
    assert p.timestamp == expected

def test_packet_v1_set_dataword():
    p = Packet_v1()
    p.dataword = 75
    expected = bah.fromuint(75, 10)
    assert p.bits[Packet_v1.dataword_bits] == expected

def test_packet_get_dataword():
    p = Packet()
    expected = 74
    p.dataword = 74
    assert p.dataword == expected

def test_packet_v1_get_dataword_ADC_bug():
    p = Packet_v1()
    expected = 74
    p.dataword = 75
    assert p.dataword == expected

def test_packet_v1_set_fifo_half_flag():
    p = Packet_v1()
    p.fifo_half_flag = 1
    expected = True
    assert p.bits[Packet_v1.fifo_half_bit] == expected

def test_packet_v1_get_fifo_half_flag():
    p = Packet_v1()
    expected = 1
    p.fifo_half_flag = expected
    assert p.fifo_half_flag == expected

def test_packet_v1_set_fifo_full_flag():
    p = Packet_v1()
    p.fifo_full_flag = 1
    expected = True
    assert p.bits[Packet_v1.fifo_full_bit] == expected

def test_packet_v1_get_fifo_full_flag():
    p = Packet_v1()
    expected = 1
    p.fifo_full_flag = expected
    assert p.fifo_full_flag == expected

def test_packet_set_register_address():
    p = Packet()
    p.register_address = 121
    if isinstance(p,Packet_v1):
        expected = bah.fromuint(121, 8)
    else:
        expected = bah.fromuint(121, 8, endian=p.endian)
    assert p.bits[Packet.register_address_bits] == expected

def test_packet_get_register_address():
    p = Packet()
    p.register_address = 18
    expected = 18
    assert p.register_address == expected

def test_packet_set_register_data():
    p = Packet()
    p.register_data = 1
    if isinstance(p,Packet_v1):
        expected = bah.fromuint(1, 8)
    else:
        expected = bah.fromuint(1, 8, endian=p.endian)
    assert p.bits[Packet.register_data_bits] == expected

def test_packet_get_register_data():
    p = Packet()
    p.register_data = 18
    expected = 18
    assert p.register_data == expected

def test_packet_v1_set_test_counter():
    p = Packet_v1()
    p.test_counter = 18376
    expected = bah.fromuint(18376, 16)
    result = (p.bits[Packet_v1.test_counter_bits_15_12] +
            p.bits[Packet_v1.test_counter_bits_11_0])
    assert result == expected

def test_packet_get_test_counter():
    p = Packet()
    expected = 19831
    p.test_counter = expected
    assert p.test_counter == expected

def test_configuration_error_on_unknown_field():
    c = Configuration()
    with pytest.raises(AttributeError):
        c.this_is_a_dummy_attr = 0
        pytest.fail('Should fail: attribute is not in known '
                       'register names')

def test_configuration_v1_no_error_on_known_register_name():
    c = Configuration_v1()
    c.reset_cycles = 5

def test_configuration_no_error_on_underscore():
    c = Configuration()
    c._underscore = 'hello'

def test_configuration_no_error_on_hasattr():
    c = Configuration()
    c.num_registers = 0

def test_configuration_get_nondefault_registers():
    c = Configuration()
    expected = {}
    assert c.get_nondefault_registers() == expected
    c.adc_burst_length += 1
    expected['adc_burst_length'] = (c.adc_burst_length, c.adc_burst_length-1)
    assert c.get_nondefault_registers() == expected

def test_configuration_get_nondefault_registers_array():
    c = Configuration()
    default = c.channel_mask[1]
    c.channel_mask[1] = int(not c.channel_mask[1])
    c.channel_mask[5] = int(not c.channel_mask[5])
    result = c.get_nondefault_registers()
    print(c.channel_mask)
    expected = {
            'channel_mask': [
                ({'index': 1, 'value': c.channel_mask[1]}, {'index': 1, 'value': default}),
                ({'index': 5, 'value': c.channel_mask[5]}, {'index': 5, 'value': default})
                ]
            }
    assert result == expected

def test_configuration_get_nondefault_registers_many_changes():
    c = Configuration()
    default = list(c.channel_mask)
    c.channel_mask[:20] = list(map(lambda x: not x, c.channel_mask[:20]))
    result = c.get_nondefault_registers()
    expected = { 'channel_mask': (c.channel_mask, default) }
    assert result == expected


def test_configuration_v1_set_pixel_trim_thresholds():
    c = Configuration_v1()
    expected = [0x05] * Configuration_v1.num_channels
    c.pixel_trim_thresholds = expected
    assert c._pixel_trim_thresholds == expected
    expected[5] = 0x10
    c.pixel_trim_thresholds[5] = 0x10
    assert c._pixel_trim_thresholds == expected

def test_configuration_v1_set_pixel_trim_thresholds_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.pixel_trim_thresholds = [0x05] * (Configuration_v1.num_channels-1)
        pytest.fail(message='Should fail: wrong num_channels')
    with pytest.raises(ValueError):
        c.pixel_trim_thresholds = [0x20] * Configuration_v1.num_channels
        pytest.fail(message='Should fail: values too large')
    with pytest.raises(ValueError,):
        c.pixel_trim_thresholds = [-10] * Configuration_v1.num_channels
        pytest.fail(message='Should fail: value negative')
    with pytest.raises(ValueError):
        c.pixel_trim_thresholds = 5
        pytest.fail(message='Should fail: wrong type')

def test_configuration_v1_get_pixel_trim_thresholds():
    c = Configuration_v1()
    expected = [0x10] * Configuration_v1.num_channels
    c._pixel_trim_thresholds = expected
    assert c.pixel_trim_thresholds == expected

def test_configuration_v1_set_global_threshold():
    c = Configuration_v1()
    expected = 0x5a
    c.global_threshold = expected
    assert c._global_threshold == expected

def test_configuration_v1_set_global_threshold_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.global_threshold = 0x100
        pytest.fail(message='Should fail: values too large')
    with pytest.raises(ValueError):
        c.global_threshold = -10
        pytest.fail(message='Should fail: value negative')
    with pytest.raises(ValueError):
        c.global_threshold = True
        pytest.fail(message='Should fail: wrong type')

def test_configuration_v1_get_global_threshold():
    c = Configuration_v1()
    expected = 0x50
    c._global_threshold = expected
    assert c.global_threshold == expected

def test_configuration_v1_set_csa_gain():
    c = Configuration_v1()
    expected = 0
    c.csa_gain = expected
    assert c._csa_gain == expected

def test_configuration_v1_set_csa_gain_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.csa_gain = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.csa_gain = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_csa_gain():
    c = Configuration_v1()
    expected = 0
    c._csa_gain = expected
    assert c.csa_gain == expected

def test_configuration_v1_set_csa_bypass():
    c = Configuration_v1()
    expected = 0
    c.csa_bypass = expected
    assert c._csa_bypass == expected

def test_configuration_v1_set_csa_bypass_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.csa_bypass = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.csa_bypass = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_csa_bypass():
    c = Configuration_v1()
    expected = 0
    c._csa_bypass = expected
    assert c.csa_bypass == expected

def test_configuration_v1_set_internal_bypass():
    c = Configuration_v1()
    expected = 0
    c.internal_bypass = expected
    assert c._internal_bypass == expected

def test_configuration_v1_set_internal_bypass_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.internal_bypass = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.internal_bypass = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_internal_bypass():
    c = Configuration_v1()
    expected = 0
    c._internal_bypass = expected
    assert c.internal_bypass == expected

def test_configuration_v1_set_csa_bypass_select():
    c = Configuration_v1()
    expected = [0x1] * Configuration_v1.num_channels
    c.csa_bypass_select = expected
    assert c._csa_bypass_select == expected
    expected[5] = 0x0
    c.csa_bypass_select[5] = expected[5]
    assert c._csa_bypass_select == expected

def test_configuration_v1_set_csa_bypass_select_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.csa_bypass_select = [0x1] * (Configuration_v1.num_channels-1)
        pytest.fail('Should fail: wrong num_channels')
    with pytest.raises(ValueError):
        c.csa_bypass_select = [0x2] * Configuration_v1.num_channels
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.csa_bypass_select = [-1] * Configuration_v1.num_channels
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.csa_bypass_select = 5
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_csa_bypass_select():
    c = Configuration_v1()
    expected = [0x1] * Configuration_v1.num_channels
    c._csa_bypass_select = expected
    assert c.csa_bypass_select == expected

def test_configuration_v1_set_csa_monitor_select():
    c = Configuration_v1()
    expected = [0x0] * Configuration_v1.num_channels
    c.csa_monitor_select = expected
    assert c._csa_monitor_select == expected
    expected[5] = 0x1
    c.csa_monitor_select[5] = expected[5]
    assert c._csa_monitor_select == expected

def test_configuration_v1_set_csa_monitor_select_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.csa_monitor_select = [0x1] * (Configuration_v1.num_channels-1)
        pytest.fail('Should fail: wrong num_channels')
    with pytest.raises(ValueError):
        c.csa_monitor_select = [0x2] * Configuration_v1.num_channels
        pytest.fail('Should fail: value too lare')
    with pytest.raises(ValueError):
        c.csa_monitor_select = [-1] * Configuration_v1.num_channels
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.csa_monitor_select = 5
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_csa_monitor_select():
    c = Configuration_v1()
    expected = [0x0] * Configuration_v1.num_channels
    c._csa_monitor_select = expected
    assert c.csa_monitor_select == expected

def test_configuration_v1_set_csa_testpulse_enable():
    c = Configuration_v1()
    expected = [0x1] * Configuration_v1.num_channels
    c.csa_testpulse_enable = expected
    assert c._csa_testpulse_enable == expected
    expected[5] = 0x0
    c.csa_testpulse_enable[5] = expected[5]
    assert c._csa_testpulse_enable == expected

def test_configuration_v1_set_csa_testpulse_enable_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.csa_testpulse_enable = [0x1] * (Configuration_v1.num_channels-1)
        pytest.fail('Should fail: wrong num_channels')
    with pytest.raises(ValueError):
        c.csa_testpulse_enable = [0x2] * Configuration_v1.num_channels
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.csa_testpulse_enable = [-1] * Configuration_v1.num_channels
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.csa_testpulse_enable = 5
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_csa_testpulse_enable():
    c = Configuration_v1()
    expected = [0x1] * Configuration_v1.num_channels
    c._csa_testpulse_enable = expected
    assert c.csa_testpulse_enable == expected

def test_configuration_v1_set_csa_testpulse_dac_amplitude():
    c = Configuration_v1()
    expected = 0x5a
    c.csa_testpulse_dac_amplitude = expected
    assert c._csa_testpulse_dac_amplitude == expected

def test_configuration_v1_set_csa_testpulse_dac_amplitude_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.csa_testpulse_dac_amplitude = 0x100
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.csa_testpulse_dac_amplitude = -10
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.csa_testpulse_dac_amplitude = True
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_csa_testpulse_dac_amplitude():
    c = Configuration_v1()
    expected = 0x50
    c._csa_testpulse_dac_amplitude = expected
    assert c.csa_testpulse_dac_amplitude == expected

def test_configuration_v1_set_test_mode():
    c = Configuration_v1()
    expected = Configuration_v1.TEST_UART
    c.test_mode = expected
    assert c._test_mode == expected

def test_configuration_v1_set_test_mode_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.test_mode = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.test_mode = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_test_mode():
    c = Configuration_v1()
    expected = Configuration_v1.TEST_FIFO
    c._test_mode = expected
    assert c.test_mode == expected

def test_configuration_v1_set_cross_trigger_mode():
    c = Configuration_v1()
    expected = 0
    c.cross_trigger_mode = expected
    assert c._cross_trigger_mode == expected

def test_configuration_v1_set_cross_trigger_mode_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.cross_trigger_mode = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.cross_trigger_mode = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_cross_trigger_mode():
    c = Configuration_v1()
    expected = 0
    c._cross_trigger_mode = expected
    assert c.cross_trigger_mode == expected

def test_configuration_v1_set_periodic_reset():
    c = Configuration_v1()
    expected = 0
    c.periodic_reset = expected
    assert c._periodic_reset == expected

def test_configuration_v1_set_periodic_reset_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.periodic_reset = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.periodic_reset = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_periodic_reset():
    c = Configuration_v1()
    expected = 0
    c._periodic_reset = expected
    assert c.periodic_reset == expected

def test_configuration_v1_set_fifo_diagnostic():
    c = Configuration_v1()
    expected = 0
    c.fifo_diagnostic = expected
    assert c._fifo_diagnostic == expected

def test_configuration_v1_set_fifo_diagnostic_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.fifo_diagnostic = 5
        pytest.fail('Should fail: invalid value')
    with pytest.raises(ValueError):
        c.fifo_diagnostic = False
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_fifo_diagnostic():
    c = Configuration_v1()
    expected = 0
    c._fifo_diagnostic = expected
    assert c.fifo_diagnostic == expected

def test_configuration_v1_set_test_burst_length():
    c = Configuration_v1()
    expected = 0x125a
    c.test_burst_length = expected
    assert c._test_burst_length == expected

def test_configuration_v1_set_test_burst_length_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.test_burst_length = 0x10000
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.test_burst_length = -10
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.test_burst_length = True
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_test_burst_length():
    c = Configuration_v1()
    expected = 0x502e
    c._test_burst_length = expected
    assert c.test_burst_length == expected

def test_configuration_v1_set_adc_burst_length():
    c = Configuration_v1()
    expected = 0x5a
    c.adc_burst_length = expected
    assert c._adc_burst_length == expected

def test_configuration_v1_set_adc_burst_length_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.adc_burst_length = 0x100
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.adc_burst_length = -10
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.adc_burst_length = True
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_adc_burst_length():
    c = Configuration_v1()
    expected = 0x50
    c._adc_burst_length = expected
    assert c.adc_burst_length == expected

def test_configuration_v1_set_channel_mask():
    c = Configuration_v1()
    expected = [0x1] * Configuration_v1.num_channels
    c.channel_mask = expected
    assert c._channel_mask == expected
    expected[5] = 0x0
    c.channel_mask[5] = expected[5]
    assert c._channel_mask == expected

def test_configuration_v1_set_channel_mask_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.channel_mask = [0x1] * (Configuration_v1.num_channels-1)
        pytest.fail('Should fail: wrong num_channels')
    with pytest.raises(ValueError):
        c.channel_mask = [0x2] * Configuration_v1.num_channels
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.channel_mask = [-1] * Configuration_v1.num_channels
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.channel_mask = 5
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_channel_mask():
    c = Configuration_v1()
    expected = [0x1] * Configuration_v1.num_channels
    c._channel_mask = expected
    assert c.channel_mask == expected

def test_configuration_v1_set_external_trigger_mask():
    c = Configuration_v1()
    expected = [0x0] * Configuration_v1.num_channels
    c.external_trigger_mask = expected
    assert c._external_trigger_mask == expected
    expected[5] = 0x1
    c.external_trigger_mask[5] = expected[5]
    assert c._external_trigger_mask == expected

def test_configuration_v1_set_external_trigger_mask_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.external_trigger_mask = [0x1] * (Configuration_v1.num_channels-1)
        pytest.fail('Should fail: wrong num_channels')
    with pytest.raises(ValueError):
        c.external_trigger_mask = [0x2] * Configuration_v1.num_channels
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.external_trigger_mask = [-1] * Configuration_v1.num_channels
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.external_trigger_mask = 5
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_external_trigger_mask():
    c = Configuration_v1()
    expected = [0x0] * Configuration_v1.num_channels
    c._external_trigger_mask = expected
    assert c.external_trigger_mask == expected

def test_configuration_v1_set_reset_cycles():
    c = Configuration_v1()
    expected = 0x125abc
    c.reset_cycles = expected
    assert c._reset_cycles == expected

def test_configuration_v1_set_reset_cycles_errors():
    c = Configuration_v1()
    with pytest.raises(ValueError):
        c.reset_cycles = 0x1000000
        pytest.fail('Should fail: value too large')
    with pytest.raises(ValueError):
        c.reset_cycles = -10
        pytest.fail('Should fail: value negative')
    with pytest.raises(ValueError):
        c.reset_cycles = True
        pytest.fail('Should fail: wrong type')

def test_configuration_v1_get_reset_cycles():
    c = Configuration_v1()
    expected = 0x502ef2
    c._reset_cycles = expected
    assert c.reset_cycles == expected

def test_configuration_v1_disable_channels():
    c = Configuration_v1()
    expected = [0, 1] * 16
    c.disable_channels(range(1, 32, 2))
    assert c.channel_mask == expected

def test_configuration_v1_disable_channels_default():
    c = Configuration_v1()
    expected = [1] * 32
    c.disable_channels()
    assert c.channel_mask == expected

def test_configuration_v1_enable_channels():
    c = Configuration_v1()
    expected = [0, 1] * 16
    c.disable_channels()
    c.enable_channels(range(0, 32, 2))
    assert c.channel_mask == expected

def test_configuration_v1_enable_channels_default():
    c = Configuration_v1()
    expected = [0] * 32
    c.disable_channels()
    c.enable_channels()
    assert c.channel_mask == expected

def test_configuration_v1_enable_external_trigger():
    c = Configuration_v1()
    expected = [0, 1] * 16
    c.enable_external_trigger(range(0, 32, 2))
    assert c.external_trigger_mask == expected

def test_configuration_v1_enable_external_trigger_default():
    c = Configuration_v1()
    expected = [0] * 32
    c.enable_external_trigger()
    assert c.external_trigger_mask == expected

def test_configuration_v1_disable_external_trigger():
    c = Configuration_v1()
    expected = [0, 1] * 16
    c.enable_external_trigger()
    c.disable_external_trigger(range(1, 32, 2))
    assert c.external_trigger_mask == expected

def test_configuration_v1_enable_testpulse():
    c = Configuration_v1()
    expected = [1, 0] * 16
    c.disable_testpulse()
    c.enable_testpulse(range(1, 32, 2))
    assert c.csa_testpulse_enable == expected

def test_configuration_v1_enable_testpulse_default():
    c = Configuration_v1()
    expected = [0] * 32
    c.enable_testpulse()
    assert c.csa_testpulse_enable == expected

def test_configuration_v1_disable_testpulse():
    c = Configuration_v1()
    expected = [1, 0] * 16
    c.enable_testpulse()
    c.disable_testpulse(range(0, 32, 2))
    assert c.csa_testpulse_enable == expected

def test_configuration_v1_disable_testpulse_default():
    c = Configuration_v1()
    expected = [1] * 32
    c.disable_testpulse()
    assert c.csa_testpulse_enable == expected

def test_configuration_v1_enable_analog_monitor():
    c = Configuration_v1()
    expected = [0, 0, 1] + [0] * 29
    c.enable_analog_monitor(2)
    assert c.csa_monitor_select == expected

def test_configuration_v1_disable_analog_monitor():
    c = Configuration_v1()
    expected = [0] * 32
    c.enable_analog_monitor(5)
    c.disable_analog_monitor()
    assert c.csa_monitor_select == expected

def test_configuration_v1_trim_threshold_data():
    c = Configuration_v1()
    expected = bah.fromuint(int('0x10', 16), 8)
    assert c.trim_threshold_data(0) == expected

def test_configuration_v1_global_threshold_data():
    c = Configuration_v1()
    expected = bah.fromuint(int('0x10', 16), 8)
    assert c.global_threshold_data() == expected

def test_configuration_v1_csa_gain_and_bypasses_data():
    c = Configuration_v1()
    expected = bitarray('00000001')
    assert c.csa_gain_and_bypasses_data() == expected

def test_configuration_v1_csa_bypass_select_data():
    c = Configuration_v1()
    c.csa_bypass_select[4] = 1
    expected = bitarray('00010000')
    assert c.csa_bypass_select_data(0) == expected
    c.csa_bypass_select[10] = 1
    expected = bitarray('00000100')
    assert c.csa_bypass_select_data(1) == expected
    c.csa_bypass_select[20] = 1
    expected = bitarray('00010000')
    assert c.csa_bypass_select_data(2) == expected
    c.csa_bypass_select[30] = 1
    expected = bitarray('01000000')
    assert c.csa_bypass_select_data(3) == expected

def test_configuration_v1_csa_monitor_select_data():
    c = Configuration_v1()
    c.csa_monitor_select[4] = 1
    expected = bitarray('00010000')
    assert c.csa_monitor_select_data(0) == expected
    c.csa_monitor_select[10] = 1
    expected = bitarray('00000100')
    assert c.csa_monitor_select_data(1) == expected
    c.csa_monitor_select[20] = 1
    expected = bitarray('00010000')
    assert c.csa_monitor_select_data(2) == expected
    c.csa_monitor_select[30] = 1
    expected = bitarray('01000000')
    assert c.csa_monitor_select_data(3) == expected

def test_configuration_v1_csa_testpulse_enable_data():
    c = Configuration_v1()
    c.csa_testpulse_enable[4] = 0
    expected = bitarray('11101111')
    assert c.csa_testpulse_enable_data(0) == expected
    c.csa_testpulse_enable[10] = 0
    expected = bitarray('11111011')
    assert c.csa_testpulse_enable_data(1) == expected
    c.csa_testpulse_enable[20] = 0
    expected = bitarray('11101111')
    assert c.csa_testpulse_enable_data(2) == expected
    c.csa_testpulse_enable[30] = 0
    expected = bitarray('10111111')
    assert c.csa_testpulse_enable_data(3) == expected

def test_configuration_v1_csa_testpulse_dac_amplitude_data():
    c = Configuration_v1()
    c.csa_testpulse_dac_amplitude = 200;
    expected = bitarray('11001000')
    assert c.csa_testpulse_dac_amplitude_data() == expected

def test_configuration_v1_test_mode_xtrig_reset_diag_data():
    c = Configuration_v1()
    c.test_mode = 2
    c.fifo_diagnostic = 1
    expected = bitarray('00010010')
    assert c.test_mode_xtrig_reset_diag_data() == expected

def test_configuration_v1_sample_cycles_data():
    c = Configuration_v1()
    c.sample_cycles = 221
    expected = bitarray('11011101')
    assert c.sample_cycles_data() == expected

def test_configuration_v1_test_burst_length_data():
    c = Configuration_v1()
    expected = bah.fromuint(int('0xFF', 16), 8)
    assert c.test_burst_length_data(0) == expected
    expected = bah.fromuint(int('0x00', 16), 8)
    assert c.test_burst_length_data(1) == expected

def test_configuration_v1_adc_burst_length_data():
    c = Configuration_v1()
    c.adc_burst_length = 140
    expected = bitarray('10001100')
    assert c.adc_burst_length_data() == expected

def test_configuration_v1_channel_mask_data():
    c = Configuration_v1()
    c.channel_mask[4] = 1
    expected = bitarray('00010000')
    assert c.channel_mask_data(0) == expected
    c.channel_mask[10] = 1
    expected = bitarray('00000100')
    assert c.channel_mask_data(1) == expected
    c.channel_mask[20] = 1
    expected = bitarray('00010000')
    assert c.channel_mask_data(2) == expected
    c.channel_mask[30] = 1
    expected = bitarray('01000000')
    assert c.channel_mask_data(3) == expected

def test_configuration_v1_external_trigger_mask_data():
    c = Configuration_v1()
    c.external_trigger_mask[4] = 0
    expected = bitarray('11101111')
    assert c.external_trigger_mask_data(0) == expected
    c.external_trigger_mask[10] = 0
    expected = bitarray('11111011')
    assert c.external_trigger_mask_data(1) == expected
    c.external_trigger_mask[20] = 0
    expected = bitarray('11101111')
    assert c.external_trigger_mask_data(2) == expected
    c.external_trigger_mask[30] = 0
    expected = bitarray('10111111')
    assert c.external_trigger_mask_data(3) == expected

def test_configuration_v1_reset_cycles_data():
    c = Configuration_v1()
    c.reset_cycles = 0xabcdef
    expected = bah.fromuint(int('0xef', 16), 8)
    assert c.reset_cycles_data(0) == expected
    expected = bah.fromuint(int('0xcd', 16), 8)
    assert c.reset_cycles_data(1) == expected
    expected = bah.fromuint(int('0xab', 16), 8)
    assert c.reset_cycles_data(2) == expected

def test_configuration_v1_to_dict():
    c = Configuration_v1()
    attrs = [
            'pixel_trim_thresholds',
            'global_threshold',
            'csa_gain',
            'csa_bypass',
            'internal_bypass',
            'csa_bypass_select',
            'csa_monitor_select',
            'csa_testpulse_enable',
            'csa_testpulse_dac_amplitude',
            'test_mode',
            'cross_trigger_mode',
            'periodic_reset',
            'fifo_diagnostic',
            'sample_cycles',
            'test_burst_length',
            'adc_burst_length',
            'channel_mask',
            'external_trigger_mask',
            'reset_cycles']
    expected = {}
    for attr in attrs:
        expected[attr] = getattr(c, attr)
    result = c.to_dict()
    assert result == expected

def test_configuration_v1_from_dict():
    c = Configuration_v1()
    attrs = [
            'pixel_trim_thresholds',
            'global_threshold',
            'csa_gain',
            'csa_bypass',
            'internal_bypass',
            'csa_bypass_select',
            'csa_monitor_select',
            'csa_testpulse_enable',
            'csa_testpulse_dac_amplitude',
            'test_mode',
            'cross_trigger_mode',
            'periodic_reset',
            'fifo_diagnostic',
            'sample_cycles',
            'test_burst_length',
            'adc_burst_length',
            'channel_mask',
            'external_trigger_mask',
            'reset_cycles']
    expected = {}
    for attr in attrs:
        expected[attr] = getattr(c, attr)
    expected['global_threshold'] = 50
    c.from_dict(expected)
    result = c.to_dict()
    assert result == expected

def test_configuration_write(tmpdir):
    c = Configuration()
    f = str(tmpdir.join('test_config.json'))
    c.write(f)
    with open(f, 'r') as output:
        result = json.load(output)
    expected = c.to_dict()
    assert result['register_values'] == expected

def test_configuration_write_errors(tmpdir):
    c = Configuration()
    f = tmpdir.join('test_config.json')
    f.write("Test data.....")
    with pytest.raises(IOError):
        c.write(str(f))
        pytest.fail('Should fail: force fails')

def test_configuration_write_force(tmpdir):
    c = Configuration()
    f = tmpdir.join('test_config.json')
    f.write("Test data.....")
    c.write(str(f), force=True)
    with open(str(f), 'r') as output:
        result = json.load(output)
    expected = c.to_dict()
    assert result['register_values'] == expected

def test_configuration_v1_read_absolute(tmpdir):
    c = Configuration_v1()
    c.pixel_trim_thresholds[0] = 30
    c.reset_cycles = 0x100010
    f = str(tmpdir.join('test_config.json'))
    c.write(f)
    c2 = Configuration_v1()
    c2.load(f)
    expected = c.to_dict()
    result = c2.to_dict()
    assert result == expected

def test_configuration_v1_read_default():
    c = Configuration_v1()
    expected = c.to_dict()
    c.global_threshold = 100
    c.load('chip/default.json')
    result = c.to_dict()
    assert result == expected

def test_configuration_v1_read_local():
    c = Configuration_v1()
    c.global_threshold = 104
    expected = c.to_dict()
    abspath = os.path.join(os.getcwd(), 'test_config.json')
    c.write(abspath)
    c.global_threshold = 0x10
    c.load('test_config.json')
    result = c.to_dict()
    os.remove(abspath)
    assert result == expected

def test_configuration_v1_from_dict_reg_pixel_trim():
    c = Configuration_v1()
    register_dict = { 0: 5, 15: 20 }
    c.from_dict_registers(register_dict)
    result_1 = c.pixel_trim_thresholds[0]
    expected_1 = register_dict[0]
    assert result_1 == expected_1
    result_2 = c.pixel_trim_thresholds[15]
    expected_2 = register_dict[15]
    assert result_2 == expected_2

def test_configuration_v1_from_dict_reg_global_threshold():
    c = Configuration_v1()
    register_dict = { 32: 182 }
    c.from_dict_registers(register_dict)
    result = c.global_threshold
    expected = register_dict[32]
    assert result == expected

def test_configuration_v1_from_dict_reg_csa_gain():
    c = Configuration_v1()
    register_dict = { 33: 0 }
    c.from_dict_registers(register_dict)
    result = c.csa_gain
    expected = 0
    assert result == expected

def test_configuration_v1_from_dict_reg_csa_bypass():
    c = Configuration_v1()
    register_dict = { 33: 2 }
    c.from_dict_registers(register_dict)
    result = c.csa_bypass
    expected = 1
    assert result == expected

def test_configuration_v1_from_dict_reg_internal_bypass():
    c = Configuration_v1()
    register_dict = { 33: 8 }
    c.from_dict_registers(register_dict)
    result = c.internal_bypass
    expected = 1
    assert result == expected

def test_configuration_v1_from_dict_reg_csa_bypass_select():
    c = Configuration_v1()
    register_dict = { 34: 0x12, 35: 0x34, 36: 0x56, 37: 0x78 }
    c.from_dict_registers(register_dict)
    result = c.csa_bypass_select
    expected = [
            0, 1, 0, 0, 1, 0, 0, 0,
            0, 0, 1, 0, 1, 1, 0, 0,
            0, 1, 1, 0, 1, 0, 1, 0,
            0, 0, 0, 1, 1, 1, 1, 0
            ]
    assert result == expected

def test_configuration_v1_from_dict_reg_csa_monitor_select():
    c = Configuration_v1()
    register_dict = { 38: 0x12, 39: 0x34, 40: 0x56, 41: 0x78 }
    c.from_dict_registers(register_dict)
    result = c.csa_monitor_select
    expected = [
            0, 1, 0, 0, 1, 0, 0, 0,
            0, 0, 1, 0, 1, 1, 0, 0,
            0, 1, 1, 0, 1, 0, 1, 0,
            0, 0, 0, 1, 1, 1, 1, 0
            ]
    assert result == expected

def test_configuration_v1_from_dict_reg_csa_testpulse_enable():
    c = Configuration_v1()
    register_dict = { 42: 0x12, 43: 0x34, 44: 0x56, 45: 0x78 }
    c.from_dict_registers(register_dict)
    result = c.csa_testpulse_enable
    expected = [
            0, 1, 0, 0, 1, 0, 0, 0,
            0, 0, 1, 0, 1, 1, 0, 0,
            0, 1, 1, 0, 1, 0, 1, 0,
            0, 0, 0, 1, 1, 1, 1, 0
            ]
    assert result == expected

def test_configuration_v1_from_dict_reg_csa_testpulse_dac_amplitude():
    c = Configuration_v1()
    register_dict = { 46: 193 }
    c.from_dict_registers(register_dict)
    result = c.csa_testpulse_dac_amplitude
    expected = 193
    assert result == expected

def test_configuration_v1_from_dict_reg_test_mode():
    c = Configuration_v1()
    register_dict = { 47: 2 }
    c.from_dict_registers(register_dict)
    result = c.test_mode
    expected = 2
    assert result == expected

def test_configuration_v1_from_dict_reg_cross_trigger_mode():
    c = Configuration_v1()
    register_dict = { 47: 4 }
    c.from_dict_registers(register_dict)
    result = c.cross_trigger_mode
    expected = 1
    assert result == expected

def test_configuration_v1_from_dict_reg_periodic_reset():
    c = Configuration_v1()
    register_dict = { 47: 8 }
    c.from_dict_registers(register_dict)
    result = c.periodic_reset
    expected = 1
    assert result == expected

def test_configuration_v1_from_dict_reg_fifo_diagnostic():
    c = Configuration_v1()
    register_dict = { 47: 16 }
    c.from_dict_registers(register_dict)
    result = c.fifo_diagnostic
    expected = 1
    assert result == expected

def test_configuration_v1_from_dict_reg_sample_cycles():
    c = Configuration_v1()
    register_dict = { 48: 111 }
    c.from_dict_registers(register_dict)
    result = c.sample_cycles
    expected = 111
    assert result == expected

def test_configuration_v1_from_dict_reg_test_burst_length():
    c = Configuration_v1()
    register_dict = { 49: 5, 50: 2}
    c.from_dict_registers(register_dict)
    result = c.test_burst_length
    expected = 517  # = 256 * 2 + 1 * 5
    assert result == expected

def test_configuration__v1from_dict_reg_adc_burst_length():
    c = Configuration_v1()
    register_dict = { 51: 83 }
    c.from_dict_registers(register_dict)
    result = c.adc_burst_length
    expected = 83
    assert result == expected

def test_configuration_v1_from_dict_reg_channel_mask():
    c = Configuration_v1()
    register_dict = { 52: 0x12, 53: 0x34, 54: 0x56, 55: 0x78 }
    c.from_dict_registers(register_dict)
    result = c.channel_mask
    expected = [
            0, 1, 0, 0, 1, 0, 0, 0,
            0, 0, 1, 0, 1, 1, 0, 0,
            0, 1, 1, 0, 1, 0, 1, 0,
            0, 0, 0, 1, 1, 1, 1, 0
            ]
    assert result == expected

def test_configuration_v1_from_dict_reg_external_trigger_mask():
    c = Configuration_v1()
    register_dict = { 56: 0x12, 57: 0x34, 58: 0x56, 59: 0x78 }
    c.from_dict_registers(register_dict)
    result = c.external_trigger_mask
    expected = [
            0, 1, 0, 0, 1, 0, 0, 0,
            0, 0, 1, 0, 1, 1, 0, 0,
            0, 1, 1, 0, 1, 0, 1, 0,
            0, 0, 0, 1, 1, 1, 1, 0
            ]
    assert result == expected

def test_configuration_v1_from_dict_reg_reset_cycles():
    c = Configuration_v1()
    register_dict = { 60: 0x12, 61: 0x34, 62: 0x56 }
    c.from_dict_registers(register_dict)
    result = c.reset_cycles
    expected = 0x563412
    assert result == expected

def test_controller_get_chip(chip):
    controller = Controller()
    chip_key = chip.chip_key
    controller.chips[chip_key] = chip
    assert controller[chip_key] == chip
    assert controller[tuple(chip_key)] == chip
    assert controller[chip_key.keystring] == chip

def test_controller_get_chip_error(chip):
    controller = Controller()
    controller.chips[chip.chip_key] = chip
    test_key_dict = chip.chip_key.to_dict()
    test_key_dict['chip_id'] += 1
    test_key = Key.from_dict(test_key_dict)
    with pytest.raises(ValueError):
        controller[test_key]
    test_key_dict = chip.chip_key.to_dict()
    test_key_dict['io_channel'] += 1
    test_key = Key.from_dict(test_key_dict)
    with pytest.raises(ValueError):
        controller[test_key]
    test_key_dict = chip.chip_key.to_dict()
    test_key_dict['io_group'] += 1
    test_key = Key.from_dict(test_key_dict)
    with pytest.raises(ValueError):
        controller[test_key]

def test_controller_read():
    controller = Controller()
    controller.io = FakeIO()
    controller.io.queue.append(([Packet()], b'\x00\x00'))
    controller.start_listening()
    result = controller.read()
    controller.stop_listening()
    expected =([Packet()], b'\x00\x00')
    assert result == expected

def test_controller_send(capfd):
    controller = Controller()
    controller.io = FakeIO()
    to_send = [Packet(b'0'*Packet.num_bytes), Packet(b'a'*Packet.num_bytes)]
    controller.send(to_send)
    result, err = capfd.readouterr()
    expected = list_of_packets_str(to_send)
    assert result == expected

def test_controller_load():
    controller = Controller()
    controller.io = FakeIO()
    controller.load('controller/pcb-1_chip_info.json')

def test_controller_read_configuration(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    conf_data = chip.get_configuration_packets(Packet.CONFIG_READ_PACKET)
    sent_expected = list_of_packets_str(conf_data)
    controller.io.queue.append((conf_data,b'hi'))
    received_expected = PacketCollection(conf_data, b'hi', read_id=0,
            message='configuration read')
    controller.read_configuration(chip.chip_key, timeout=0.01)
    received_result = controller.reads[-1]
    sent_result, err = capfd.readouterr()
    assert sent_result == sent_expected
    assert received_result == received_expected

def test_controller_read_configuration_reg(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    conf_data = chip.get_configuration_packets(Packet.CONFIG_READ_PACKET)[0:1]
    sent_expected = list_of_packets_str(conf_data)
    controller.io.queue.append((conf_data,b'hi'))
    received_expected = PacketCollection(conf_data, b'hi', read_id=0,
            message='configuration read')
    controller.read_configuration(chip.chip_key, 0, timeout=0.01)
    received_result = controller.reads[-1]
    sent_result, err = capfd.readouterr()
    assert sent_result == sent_expected
    assert received_result == received_expected

def test_controller_write_configuration(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    controller.write_configuration(chip.chip_key)
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    expected = list_of_packets_str(conf_data)
    result, err = capfd.readouterr()
    assert result == expected

def test_controller_write_configuration_one_reg(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    controller.write_configuration(chip.chip_key, 0)
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)[0:1]
    expected = list_of_packets_str(conf_data)
    result, err = capfd.readouterr()
    assert result == expected

def test_controller_write_configuration_write_read(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    to_read = ([Packet(b'1'*Packet.num_bytes)], b'hi')
    controller.io.queue.append(to_read)
    expected_read = PacketCollection(*to_read, read_id=0,
            message='configuration write')
    controller.write_configuration(chip.chip_key, registers=5, write_read=0.1)
    result_read = controller.reads[0]
    assert result_read == expected_read
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)[5:6]
    expected_sent = list_of_packets_str(conf_data)
    result_sent, err = capfd.readouterr()
    assert result_sent == expected_sent

def test_controller_multi_write_configuration(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    key = chip.chip_key
    key2_dict = chip.chip_key.to_dict()
    key2_dict['chip_id'] += 1
    key2 = Key.from_dict(key2_dict)
    chip2 = Chip(key2)
    controller.chips = {key:chip, key2:chip2}
    controller.multi_write_configuration((key, key2))
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    conf_data2 = chip2.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    final_string_1 = list_of_packets_str(conf_data)
    final_string_2 = list_of_packets_str(conf_data2)
    expected = final_string_1 + final_string_2
    result, err = capfd.readouterr()
    assert result == expected

def test_controller_multi_write_configuration_write_read(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    to_read = ([Packet(b'1'*Packet.num_bytes)], b'hi')
    controller.io.queue.append(to_read)
    expected_read = PacketCollection(*to_read, read_id=0,
            message='configuration write')
    key = chip.chip_key
    key2_dict = chip.chip_key.to_dict()
    key2_dict['chip_id'] += 1
    key2 = Key.from_dict(key2_dict)
    chip2 = Chip(key2)
    controller.chips = {key:chip, key2:chip2}
    controller.multi_write_configuration((key, key2), write_read=0.01)
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    conf_data2 = chip2.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    final_string_1 = list_of_packets_str(conf_data)
    final_string_2 = list_of_packets_str(conf_data2)
    expected = final_string_1 + final_string_2
    result, err = capfd.readouterr()
    assert result == expected

def test_controller_multi_write_configuration_specify_registers(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    key = chip.chip_key
    key2_dict = chip.chip_key.to_dict()
    key2_dict['chip_id'] += 1
    key2 = Key.from_dict(key2_dict)
    chip2 = Chip(key2)
    controller.chips = {key:chip, key2:chip2}
    controller.multi_write_configuration([(key, 0), key2])
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)[:1]
    conf_data2 = chip2.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    final_string_1 = list_of_packets_str(conf_data)
    final_string_2 = list_of_packets_str(conf_data2)
    expected = final_string_1 + final_string_2
    result, err = capfd.readouterr()
    assert result == expected


def test_controller_multi_read_configuration(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.use_all_chips = True
    key0_dict = chip.chip_key.to_dict()
    key0_dict['io_channel'] = 1
    key0_dict['chip_key'] = 1
    key0 = Key.from_dict(key0_dict)
    key1_dict = chip.chip_key.to_dict()
    key1_dict['chip_id'] += 1
    key1_dict['io_channel'] = 1
    key1_dict['chip_key'] = 1
    key1 = Key.from_dict(key1_dict)
    chip1 = controller.add_chip(key0)
    chip2 = controller.add_chip(key1)
    conf_data = chip1.get_configuration_packets(Packet.CONFIG_READ_PACKET)
    conf_data2 = chip2.get_configuration_packets(Packet.CONFIG_READ_PACKET)
    final_string_1 = list_of_packets_str(conf_data)
    final_string_2 = list_of_packets_str(conf_data2)
    expected_sent = final_string_1 + final_string_2
    controller.io.queue.append((conf_data+conf_data2, b'hi'))
    expected_read = PacketCollection(conf_data+conf_data2, b'hi',
            read_id=0, message='multi configuration read')
    controller.multi_read_configuration((key0, key1), timeout=0.01)
    result_sent, err = capfd.readouterr()
    result_read = controller.reads[-1]
    assert result_sent == expected_sent
    assert result_read == expected_read

def test_controller_multi_read_configuration_specify_registers(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    key0 = Key(chip.chip_key)
    key1_dict = chip.chip_key.to_dict()
    key1_dict['chip_id'] += 1
    key1 = Key.from_dict(key1_dict)
    controller.chips[key0] = Chip(chip_key=key0)
    controller.chips[key1] = Chip(chip_key=key1)
    conf_data = controller.chips[key0].get_configuration_packets(Packet.CONFIG_READ_PACKET)[:1]
    conf_data2 = controller.chips[key1].get_configuration_packets(Packet.CONFIG_READ_PACKET)
    final_string_1 = list_of_packets_str(conf_data)
    final_string_2 = list_of_packets_str(conf_data2)
    expected_sent = final_string_1 + final_string_2
    controller.io.queue.append((conf_data+conf_data2, b'hi'))
    expected_read = PacketCollection(conf_data+conf_data2, b'hi',
            read_id=0, message='multi configuration read')
    controller.multi_read_configuration([(key0, 0), key1], timeout=0.01)
    result_sent, err = capfd.readouterr()
    result_read = controller.reads[-1]
    assert result_sent == expected_sent
    assert result_read == expected_read

def test_controller_verify_configuration_ok(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    for packet in conf_data: packet.packet_type = Packet.CONFIG_READ_PACKET
    controller.io.queue.append((conf_data,b'hi'))
    ok, diff = controller.verify_configuration(chip_keys=chip.chip_key)
    assert diff == {}
    assert ok

def test_controller_verify_configuration_missing_packet(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    for packet in conf_data: packet.packet_type = Packet.CONFIG_READ_PACKET
    del conf_data[5]
    controller.io.queue.append((conf_data,b'hi'))
    ok, diff = controller.verify_configuration(chip_keys=chip.chip_key)
    assert ok == False
    assert diff == {chip.chip_key: {5: (16, None)}}

def test_controller_verify_configuration_bad_value(capfd, chip):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip.chip_key] = chip
    conf_data = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    for packet in conf_data: packet.packet_type = Packet.CONFIG_READ_PACKET
    conf_data[5].register_data = 17
    controller.io.queue.append((conf_data,b'hi'))
    ok, diff = controller.verify_configuration(chip_keys=chip.chip_key)
    assert ok == False
    assert diff == {chip.chip_key: {5: (16, 17)}}

def test_packetcollection_getitem_int():
    expected = Packet()
    collection = PacketCollection([expected])
    result = collection[0]
    assert result == expected

def test_packetcollection_getitem_int_bits(timestamp_packet):
    packet = Packet()
    collection = PacketCollection([packet])
    result = collection[0, 'bits']
    expected = ' '.join(packet.bits.to01()[i:i+8] for i in range(0, Packet.size, 8))
    assert result == expected
    packet2 = timestamp_packet
    collection2 = PacketCollection([packet2])
    result2 = collection2[0, 'bits']
    expected2 = ' '.join(packet2.bits.to01()[i:i+8] for i in range(0,
        packet2.size, 8))
    assert result2 == expected2

def test_packetcollection_getitem_slice(chip):
    packets = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    collection = PacketCollection(packets, message='hello')
    result = collection[:10]
    expected = PacketCollection(packets[:10], message='hello'
        ' | subset slice(None, 10, None)')
    assert result == expected

def test_packetcollection_getitem_slice_bits(chip):
    packets = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    collection = PacketCollection(packets, message='hello')
    result = collection[:10, 'bits']
    expected = [' '.join(p.bits.to01()[i:i+8] for i in range(0,
        Packet.size, 8)) for p in packets[:10]]
    assert result == expected

def test_packetcollection_origin(chip):
    packets = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)
    collection = PacketCollection(packets, message='hello')
    first_gen = collection.by_chipid()[chip.chip_id]
    second_gen = first_gen.by_chipid()[chip.chip_id]
    assert first_gen.parent is collection
    assert first_gen.origin() is collection
    assert second_gen.parent is first_gen
    assert second_gen.origin() is collection

def test_packetcollection_v2_extract():
    p1 = Packet_v2()
    p1.chip_id = 10
    p1.packet_type = Packet_v2.DATA_PACKET
    p1.dataword = 36
    p2 = Packet_v2()
    p2.chip_id = 9
    p2.packet_type = Packet_v2.DATA_PACKET
    p2.dataword = 38
    p3 = Packet_v2()
    p3.chip_id = 8
    p3.packet_type = Packet_v2.TEST_PACKET
    pc = PacketCollection([p1,p2,p3])
    expected = [10, 9, 8]
    assert pc.extract('chip_id') == expected
    expected = [36, 38]
    assert pc.extract('dataword', packet_type=Packet_v2.DATA_PACKET) == expected
    expected = [36]
    assert pc.extract('dataword', chip_id=10, packet_type=Packet_v2.DATA_PACKET) == expected
    expected = [[10,36],[9,38]]
    assert pc.extract('chip_id','dataword', packet_type=Packet_v2.DATA_PACKET) == expected

def test_packetcollection_to_dict():
    packet = Packet()
    packet.chip_id = 246
    packet.packet_type = Packet.TEST_PACKET
    collection = PacketCollection([packet], bytestream=packet.bytes(),
            message='hello')
    result = collection.to_dict()
    expected = {
            'id': id(collection),
            'parent': 'None',
            'message': 'hello',
            'read_id': 'None',
            'bytestream': packet.bytes().decode('raw_unicode_escape'),
            'packets': [packet.export()]
            }
    assert result == expected

def test_packetcollection_from_dict():
    packet = Packet()
    packet.packet_type = Packet.TEST_PACKET
    collection = PacketCollection([packet], bytestream=packet.bytes(),
            message='hello')
    d = collection.to_dict()
    result = PacketCollection([])
    result.from_dict(d)
    expected = collection
    assert result == expected

def test_Smart_List_init_wrong_type():
    with pytest.raises(ValueError):
        sl = _Smart_List(5, 0, 40)
        pytest.fail('Should fail: wrong type')

def test_Smart_List_init_out_of_bounds():
    with pytest.raises(ValueError):
        sl = _Smart_List([-1], 0, 40)
        pytest.fail('Should fail: out of bounds')

def test_Smart_List_assignment():
    result = _Smart_List([1,2,3],0,40)
    expected = [1,2,3]
    assert result == expected
    result[0] = 20
    expected = [20,2,3]
    assert result == expected

def test_Smart_List_error():
    sl = _Smart_List([1,2,3],0,40)
    with pytest.raises(ValueError):
        sl[0] = 41
        pytest.fail('Should fail: out of bounds')

def test_Smart_List_slice_error():
    sl = _Smart_List(list(range(10)), 0, 40)
    with pytest.raises(ValueError):
        sl[5:7] = [40, 41, 40]
        pytest.fail('Should fail: out of bounds')

def test_Smart_List_config_error():
    c = Configuration()
    register_dict = { 0: 5, 15: 89 }
    with pytest.raises(ValueError):
        c.from_dict_registers(register_dict)
        pytest.fail('Should fail: out of bounds')

def test_ts_packet_to_dict(timestamp_packet):
    packet_dict = {
            'bits': timestamp_packet.bits.to01(),
            'type_str': 'timestamp',
            'type': timestamp_packet.packet_type,
            'timestamp': timestamp_packet.timestamp,
            }
    assert timestamp_packet.export() == packet_dict

def test_ts_packet_from_dict(timestamp_packet):
    p1 = TimestampPacket()
    packet_dict = {
            'bits': timestamp_packet.bits.to01(),
            'type_str': 'timestamp',
            'type': timestamp_packet.packet_type,
            'timestamp': timestamp_packet.timestamp,
            }
    p1.from_dict(packet_dict)
    assert timestamp_packet == p1

def test_ts_packet_from_dict_export_inv(timestamp_packet):
    p1 = TimestampPacket()
    p1.from_dict(timestamp_packet.export())
    assert timestamp_packet == p1

def test_message_packet_to_dict(message_packet):
    packet_dict = {
            'bits': message_packet.bits.to01(),
            'type_str': 'message',
            'message': message_packet.message,
            'type': message_packet.packet_type,
            'timestamp': message_packet.timestamp,
            }
    assert message_packet.export() == packet_dict

def test_message_packet_from_dict(message_packet):
    p1 = MessagePacket(None,None)
    packet_dict = {
            'bits': message_packet.bits.to01(),
            'type_str': 'message',
            'message': message_packet.message,
            'type': message_packet.packet_type,
            'timestamp': message_packet.timestamp,
            }
    p1.from_dict(packet_dict)
    assert message_packet == p1

def test_message_packet_from_dict_export_inv(message_packet):
    p1 = MessagePacket(None,None)
    p1.from_dict(message_packet.export())
    assert message_packet == p1

def test_key():
    with pytest.raises(ValueError):
        k = Key('0.0.0')
        pytest.fail('key is not proper format')
    with pytest.raises(ValueError):
        k = Key('256-0-0')
        pytest.fail('key value is not 1-byte')
    with pytest.raises(ValueError):
        k = Key(256,0,0)
        pytest.fail('key value is not 1-byte')
    with pytest.raises(TypeError):
        k = Key(250,0,0,1)
        pytest.fail('too many args')

    k = Key('1-2-3')
    with pytest.raises(AttributeError):
        k.io_channel = 1
        pytest.fail('key is pseudo immutable')
    with pytest.raises(AttributeError):
        k.io_group = 1
        pytest.fail('key is pseudo immutable')
    with pytest.raises(AttributeError):
        k.chip_id = 1
        pytest.fail('key is pseudo immutable')

    assert k == '1-2-3'
    assert Key(k) == k
    assert k.is_valid_keystring('0-0-256') == False
    assert k.is_valid_keystring('0-256-0') == False
    assert k.is_valid_keystring('256-0-0') == False
    assert k.is_valid_keystring('0 0 0') == False
    assert k.to_dict() == { 'io_group':1, 'io_channel':2, 'chip_id':3}
    test_dict = {'io_group':3, 'io_channel':2, 'chip_id':1}
    assert k.from_dict(test_dict) == Key('3-2-1')

    # test that you can hash a dict with the keys
    d = {}
    d[k] = 'test'
    assert d[k] == 'test'



'''
This module contains chip configuration files. They are formatted as standard
JSON files with the following fields: ``"_config_type"``, ``"class"``, and
``"register_values"``. The ``_config_type`` field should always be set to
``"chip"`` and is used for validation when loading the configuration. An
optional ``"_include"`` field can be used to specify other configurations to
inherit configuration values from. Configuration inheritance occurs in the order
specified in the list so if a conflict is encountered, fields from a file later
in the list will overwrite fields from a file earlier in the list.

Additionally, the ``class`` field provides additional validation. There are two
configuration classes: ``"Configuration_v1"``, ``"Configuration_v2"``, and ``"Configuration_Lightpix_v1"``, each
corresponding to its respective ``larpix.configuration`` class. Attempting to
load a ``Configuration_v1`` config file to a ``Configuration_v2`` object and
vice versa will raise a runtime error.

The ``register_values`` field is a dict of the register name and the value to
use. All non-default fields should be specified, but it is recommended to use no
more than the minimum required. The following is an example of a partially-specified v2 configuration
file, which will maintain the current configuration except for the specified fields::

      {
            "_config_type": "chip",
            "_include": []
            "class": "Configuration_v2",
            "register_values": {
                  "enable_periodic_reset": 1,
                  "enable_periodic_trigger": 1,
                  "enable_rolling_periodic_trigger": 1,
                  "periodic_trigger_cycles": 100000
            }
      }

Here is an example using the configuration file inheritance::

      {
            "_config_type": "chip",
            "_include": ["chip/default_v2.json"]
      }

Which will be expanded to be the equivalent of::

      {
            "_config_type": "chip",
            "_include": [],
            "class": "Configuration_v2",
            "register_values": {
                  "pixel_trim_dac": [16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16],
                  "threshold_global": 255,
                  "csa_gain": 1,
                  "csa_bypass_enable": 0,
                  "bypass_caps_en": 1,
                  "csa_enable": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                  "ibias_tdac": 8,
                  "ibias_comp": 8,
                  "ibias_buffer": 8,
                  "ibias_csa": 8,
                  "ibias_vref_buffer": 8,
                  "ibias_vcm_buffer": 8,
                  "ibias_tpulse": 6,
                  "ref_current_trim": 16,
                  "override_ref": 0,
                  "ref_kickstart": 0,
                  "vref_dac": 219,
                  "vcm_dac": 77,
                  "csa_bypass_select": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  "csa_monitor_select": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  "csa_testpulse_enable": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                  "csa_testpulse_dac": 0,
                  "current_monitor_bank0": [0, 0, 0, 0],
                  "current_monitor_bank1": [0, 0, 0, 0],
                  "current_monitor_bank2": [0, 0, 0, 0],
                  "current_monitor_bank3": [0, 0, 0, 0],
                  "voltage_monitor_bank0": [0, 0, 0],
                  "voltage_monitor_bank1": [0, 0, 0],
                  "voltage_monitor_bank2": [0, 0, 0],
                  "voltage_monitor_bank3": [0, 0, 0],
                  "voltage_monitor_refgen": [0, 0, 0, 0, 0, 0, 0, 0],
                  "digital_monitor_enable": 1,
                  "digital_monitor_select": 0,
                  "digital_monitor_chan": 0,
                  "slope_control0": 0,
                  "slope_control1": 0,
                  "slope_control2": 0,
                  "slope_control3": 0,
                  "chip_id": 1,
                  "load_config_defaults": 0,
                  "enable_fifo_diagnostics": 0,
                  "clk_ctrl": 0,
                  "enable_miso_upstream": [0, 0, 0, 0],
                  "enable_miso_downstream": [0, 0, 0, 0],
                  "enable_miso_differential": [0, 0, 0, 0],
                  "enable_mosi": [0, 0, 0, 0],
                  "test_mode_uart0": 0,
                  "test_mode_uart1": 0,
                  "test_mode_uart2": 0,
                  "test_mode_uart3": 0,
                  "enable_cross_trigger": 0,
                  "enable_periodic_reset": 0,
                  "enable_rolling_periodic_reset": 0,
                  "enable_periodic_trigger": 0,
                  "enable_rolling_periodic_trigger": 0,
                  "enable_periodic_trigger_veto": 1,
                  "enable_hit_veto": 1,
                  "adc_hold_delay": 0,
                  "adc_burst_length": 0,
                  "channel_mask": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                  "external_trigger_mask": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                  "cross_trigger_mask": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                  "periodic_trigger_mask": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                  "periodic_reset_cycles": 4096,
                  "periodic_trigger_cycles": 0,
                  "enable_dynamic_reset": 0,
                  "enable_min_delta_adc": 0,
                  "threshold_parity": 1,
                  "reset_length": 1,
                  "mark_first_packet": 1,
                  "reset_threshold": 0,
                  "min_delta_adc": 0,
                  "digital_threshold": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            }
      }

'''
pass



#!/usr/env/bin python3
'''
Usage
=====
To merge larpix packet-formatted hdf5 files together::

    packethdf5_tool.py --merge -i <files to merge, in order> -o <destination filename>

> Note:
>
> When merging files, there is the possibility of losing some file meta data. The
> merge behavior will only keep the metadata of each dataset coming from the first
> file. It is highly recommended that you only `merge` files where the metadata
> guaranteed to be the same (i.e. many files of the same simulation run).

'''
import h5py
import warnings
import os
try:
    from tqdm import tqdm
    _has_tqdm = True
except Exception as e:
    warnings.warn(str(e), RuntimeWarning)
    _has_tqdm = False

_default_max_length = -1
_default_block_size = 102400

def move_dataset(input_file, output_file, dset_name, block_size):
    curr_idx = len(output_file[dset_name])

    mc_assn_flag =  dset_name == 'mc_packets_assn'
    mc_offset = curr_idx


    output_file[dset_name].resize((len(output_file[dset_name]) + len(input_file[dset_name]),))
    # copy data in chunks
    for start in tqdm(range(0, len(input_file[dset_name]), block_size)) if _has_tqdm else range(0, len(input_file[dset_name]), block_size):
        end = min(start+block_size, len(input_file[dset_name]))
        prev_idx = curr_idx
        curr_idx += end - start

        if mc_assn_flag:
            # special case for mc packets associations, add offset since start of dataset to track_ids
            data = input_file[dset_name][start:end]
            data['track_ids'] += mc_offset
            output_file[dset_name][prev_idx:curr_idx] = data
        else:
            output_file[dset_name][prev_idx:curr_idx] = input_file[dset_name][start:end]

def merge_files(input_filenames, output_filename, block_size):
    with h5py.File(output_filename, 'w') as fo:
        for i,input_filename in enumerate(input_filenames):
            print(input_filename, '{}/{}'.format(i+1, len(input_filenames)))
            with h5py.File(input_filename, 'r') as fi:
                if i == 0:
                    # create datasets and groups
                    for grp_name in fi.keys():
                        if isinstance(fi[grp_name], h5py.Group):
                            fo.copy(fi[grp_name], grp_name)

                    # create datasets
                    for dset_name in fi.keys():
                        if isinstance(fi[dset_name], h5py.Dataset):
                            fo.create_dataset(dset_name, shape=(0,), maxshape=(None,), compression='gzip', dtype=fi[dset_name].dtype)

                        # copy meta data
                        for attr,value in fi[dset_name].attrs.items():
                            fo[dset_name].attrs[attr] = value

                # copy data
                for dset_name in fo.keys():
                    if isinstance(fo[dset_name], h5py.Dataset):
                        print('copying',dset_name,'...')
                        move_dataset(fi, fo, dset_name, block_size)


def main(input_filenames, output_filename, max_length=_default_max_length, block_size=_default_block_size, **kwargs):
    if kwargs.get('merge', False):
        merge_files(
            input_filenames, output_filename,
            block_size=block_size
            )
    else:
        print('No action specified, exiting.')

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-i', nargs='+', required=True, help='Input file(s)')
    parser.add_argument('-o', required=True, type=str, help='Output file')
    parser.add_argument('--block_size', type=int, default=_default_block_size, required=False, help='Block size used for reads (default=%(default)s)')
    parser.add_argument('--merge', action='store_true', help='Flag to merge files')
    args = parser.parse_args()
    main(
        input_filenames=args.i,
        output_filename=args.o,
        **vars(args)
        )
